import typer
from typing import Optional, Annotated


def install_ssh_server() -> None:
    """📡 Install SSH server"""
    import platform
    if platform.system() == "Windows":
        script = _get_windows_ssh_server_install_script()
    elif platform.system() == "Linux" or platform.system() == "Darwin":
        script = """
sudo nala install openssh-server -y || true  # try to install first
# sudo nala purge openssh-server -y
# sudo nala install openssh-server -y
echo "✅ FINISHED installing openssh-server."
"""
    else:
        raise NotImplementedError(f"Platform {platform.system()} is not supported.")
    from machineconfig.utils.code import run_shell_script
    run_shell_script(script=script)


def change_ssh_port(port: Annotated[int, typer.Option(..., "--port", "-p", help="SSH port to use")] = 2222) -> None:
    """🔌 Change SSH port (Linux/WSL only, default: 2222)"""
    import platform
    if platform.system() != "Linux":
        raise NotImplementedError("change_ssh_port requires Linux environment")
    from machineconfig.utils.ssh_utils.wsl import change_ssh_port as _change_ssh_port
    _change_ssh_port(port=port)


def _get_windows_ssh_server_install_script(use_winget: bool = True) -> str:
    install_winget = r'''
    Write-Host "Installing OpenSSH via winget..." -ForegroundColor Cyan
    winget install --no-upgrade --Id Microsoft.OpenSSH.Preview --source winget --accept-package-agreements --accept-source-agreements
    $wingetSshPath = "C:\Program Files\OpenSSH"
    $currentMachinePath = [System.Environment]::GetEnvironmentVariable("Path", "Machine")
    if ($currentMachinePath -notlike "*$wingetSshPath*") {
        Write-Host "Adding OpenSSH to system PATH..." -ForegroundColor Yellow
        [System.Environment]::SetEnvironmentVariable("Path", "$currentMachinePath;$wingetSshPath", "Machine")
    }
    $env:Path = [System.Environment]::GetEnvironmentVariable("Path", "Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path", "User")
'''
    install_capability = r'''
    Write-Host "Installing OpenSSH via Windows Capability..." -ForegroundColor Cyan
    Add-WindowsCapability -Online -Name OpenSSH.Server
    Add-WindowsCapability -Online -Name OpenSSH.Client
    $env:Path = [System.Environment]::GetEnvironmentVariable("Path", "Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path", "User")
'''
    configure_sshd_service = r'''
    $serviceExists = Get-Service -Name sshd -ErrorAction SilentlyContinue
    if (-not $serviceExists) {
        Write-Host "sshd service not found, registering manually..." -ForegroundColor Yellow
        sc.exe create sshd binPath= "$sshdExePath" start= auto
        if ($LASTEXITCODE -ne 0) { throw "Failed to create sshd service" }
    }
    Set-Service -Name sshd -StartupType Automatic
    Start-Service sshd
    Write-Host "sshd service started successfully." -ForegroundColor Green
'''
    configure_firewall = r'''
    $existingRule = Get-NetFirewallRule -Name "OpenSSH-Server-In-TCP" -ErrorAction SilentlyContinue
    if (-not $existingRule) {
        Write-Host "Creating firewall rule for SSH..." -ForegroundColor Yellow
        New-NetFirewallRule -Name "OpenSSH-Server-In-TCP" -DisplayName "OpenSSH SSH Server" -Enabled True -Direction Inbound -Protocol TCP -Action Allow -LocalPort 22 -Program $sshdExePath
    } else {
        Write-Host "Firewall rule already exists." -ForegroundColor Green
    }
'''
    configure_default_shell = r'''
    Write-Host "Configuring default shell for SSH..." -ForegroundColor Cyan
    $shell = "C:\Windows\System32\WindowsPowerShell\v1.0\powershell.exe"
    if (-not (Test-Path "HKLM:\SOFTWARE\OpenSSH")) { New-Item -Path "HKLM:\SOFTWARE\OpenSSH" -Force | Out-Null }
    New-ItemProperty -Path "HKLM:\SOFTWARE\OpenSSH" -Name DefaultShell -Value $shell -PropertyType String -Force | Out-Null
    New-ItemProperty -Path "HKLM:\SOFTWARE\OpenSSH" -Name DefaultShellCommandOption -Value "/c" -PropertyType String -Force | Out-Null
    Write-Host "Default shell set to: $shell" -ForegroundColor Green
'''
    configure_ssh_agent = r'''
    Write-Host "Configuring ssh-agent service..." -ForegroundColor Cyan
    $agentService = Get-Service -Name ssh-agent -ErrorAction SilentlyContinue
    if ($agentService) {
        Set-Service -Name ssh-agent -StartupType Automatic
        Start-Service ssh-agent -ErrorAction SilentlyContinue
        Write-Host "ssh-agent service configured and started." -ForegroundColor Green
    } else {
        Write-Host "ssh-agent service not available." -ForegroundColor Yellow
    }
'''
    ensure_ssh_directory = r'''
    $sshDir = Join-Path $env:USERPROFILE ".ssh"
    if (-not (Test-Path $sshDir)) {
        New-Item -ItemType Directory -Path $sshDir -Force | Out-Null
        Write-Host "Created .ssh directory at: $sshDir" -ForegroundColor Green
    }
'''
    if use_winget:
        install_method = install_winget
        sshd_exe_path = r'C:\Program Files\OpenSSH\sshd.exe'
    else:
        install_method = install_capability
        sshd_exe_path = r'$env:WINDIR\System32\OpenSSH\sshd.exe'
    return fr'''
$ErrorActionPreference = "Stop"
try {{ Set-ExecutionPolicy -Scope CurrentUser Bypass -ErrorAction SilentlyContinue }} catch {{ }}
Write-Host "========================================" -ForegroundColor Magenta
Write-Host "OpenSSH Server Installation for Windows" -ForegroundColor Magenta
Write-Host "========================================" -ForegroundColor Magenta
{install_method}
$sshdExePath = "{sshd_exe_path}"
{configure_sshd_service}
{configure_firewall}
{configure_default_shell}
{configure_ssh_agent}
{ensure_ssh_directory}
Write-Host "========================================" -ForegroundColor Magenta
Write-Host "OpenSSH Server installation complete!" -ForegroundColor Green
Write-Host "SSH config directory: $env:ProgramData\ssh" -ForegroundColor Cyan
Write-Host "========================================" -ForegroundColor Magenta
'''


def add_ssh_key(
    path: Annotated[Optional[str], typer.Option(..., help="Path to the public key file")] = None,
    choose: Annotated[bool, typer.Option(..., "--choose", "-c", help="Choose from available public keys in ~/.ssh/*.pub")] = False,
    value: Annotated[bool, typer.Option(..., "--value", "-v", help="Paste the public key content manually")] = False,
    github: Annotated[Optional[str], typer.Option(..., "--github", "-g", help="Fetch public keys from a GitHub username")] = None,
) -> None:
    """🔑 Add SSH public key to this machine so its accessible by owner of corresponding private key."""
    import machineconfig.scripts.python.helpers.helpers_network.ssh_add_ssh_key as helper
    helper.main(pub_path=path, pub_choose=choose, pub_val=value, from_github=github)


def add_ssh_identity() -> None:
    """🗝️ Add SSH identity (private key) to this machine"""
    import machineconfig.scripts.python.helpers.helpers_network.ssh_add_identity as helper
    helper.main()


def debug_ssh() -> None:
    """🐛 Debug SSH connection"""
    from platform import system
    if system() == "Linux" or system() == "Darwin":
        import machineconfig.scripts.python.helpers.helpers_network.ssh_debug_linux as helper
        helper.ssh_debug_linux()
    elif system() == "Windows":
        import machineconfig.scripts.python.helpers.helpers_network.ssh_debug_windows as helper
        helper.ssh_debug_windows()
    else:
        raise NotImplementedError(f"Platform {system()} is not supported.")


def get_app() -> typer.Typer:
    ssh_app = typer.Typer(help="🔐 SSH subcommands", no_args_is_help=True, add_help_option=True, add_completion=False)
    ssh_app.command(name="install-server", help="📡 [i] Install SSH server")(install_ssh_server)
    ssh_app.command(name="i", help="Install SSH server", hidden=True)(install_ssh_server)
    ssh_app.command(name="change-port", help="🔌 [p] Change SSH port (Linux/WSL only)")(change_ssh_port)
    ssh_app.command(name="p", help="Change SSH port", hidden=True)(change_ssh_port)
    ssh_app.command(name="add-key", help="🔑 [k] Add SSH public key to this machine", no_args_is_help=True)(add_ssh_key)
    ssh_app.command(name="k", help="Add SSH public key to this machine", hidden=True, no_args_is_help=True)(add_ssh_key)
    ssh_app.command(name="add-identity", help="🗝️ [A] Add SSH identity (private key) to this machine")(add_ssh_identity)
    ssh_app.command(name="A", help="Add SSH identity (private key) to this machine", hidden=True)(add_ssh_identity)
    ssh_app.command(name="debug", help="🐛 [d] Debug SSH connection")(debug_ssh)
    ssh_app.command(name="d", help="Debug SSH connection", hidden=True)(debug_ssh)
    return ssh_app




import typer
from typing import Optional, Literal, Annotated, TypeAlias

ON_CONFLICT_LOOSE: TypeAlias = Literal[
    "throw-error", "t",
    "overwrite-self-managed", "os",
    "backup-self-managed", "bs",
    "overwrite-default-path", "od",
    "backup-default-path", "bd"
    ]
ON_CONFLICT_STRICT: TypeAlias = Literal["throw-error", "overwrite-self-managed", "backup-self-managed", "overwrite-default-path", "backup-default-path"]
ON_CONFLICT_MAPPER: dict[str, ON_CONFLICT_STRICT] = {
    "t": "throw-error",
    "os": "overwrite-self-managed",
    "bs": "backup-self-managed",
    "od": "overwrite-default-path",
    "bd": "backup-default-path",
    "throw-error": "throw-error",
    "overwrite-self-managed": "overwrite-self-managed",
    "backup-self-managed": "backup-self-managed",
    "overwrite-default-path": "overwrite-default-path",
    "backup-default-path": "backup-default-path",
    }


def main_public_from_parser(method: Annotated[Literal["symlink", "s", "copy", "c"], typer.Option(..., "--method", "-m", help="Method to use for setting up the config file.")],
                            on_conflict: Annotated[ON_CONFLICT_LOOSE, typer.Option(..., "--on-conflict", "-o", help="Action to take on conflict")] = "throw-error",
                            which: Annotated[Optional[str], typer.Option(..., "--which", "-w", help="Specific items to process")] = "all",
                            interactive: Annotated[bool, typer.Option(..., "--interactive", "-i", help="Run in interactive mode")] = False):
    """Terminology:
    SOURCE = Self-Managed-Config-File-Path
    TARGET = Config-File-Default-Path
    For public config files, the source always exists, because we know it comes from machineconfig repo."""
    from machineconfig.profile.create_links import ConfigMapper, read_mapper
    mapper_full = read_mapper()["public"]
    if which is None:
        assert interactive is True
        from machineconfig.utils.options import choose_from_options
        items_chosen = choose_from_options(msg="Which symlink to create?", options=list(mapper_full.keys()), tv=True, multi=True)
    else:
        assert interactive is False
        if which == "all":
            items_chosen = list(mapper_full.keys())
        else:
            items_chosen = which.split(",")
    items_objections: dict[str, list[ConfigMapper]] = {item: mapper_full[item] for item in items_chosen if item in mapper_full}
    if len(items_objections) == 0:
        typer.echo("[red]Error:[/] No valid items selected.")
        typer.Exit(code=1)
        return
    from machineconfig.profile.create_links import apply_mapper
    from machineconfig.profile.create_helper import copy_assets_to_machine
    copy_assets_to_machine(which="settings")  # config files live here and will be linked to.
    method_map: dict[str, Literal["symlink", "copy"]] = {
        "s": "symlink",
        "symlink": "symlink",
        "c": "copy",
        "copy": "copy",
    }
    method = method_map[method]
    apply_mapper(mapper_data=items_objections, on_conflict=ON_CONFLICT_MAPPER[on_conflict], method=method)


def main_private_from_parser(method: Annotated[Literal["symlink", "s", "copy", "c"], typer.Option(..., "--method", "-m", help="Method to use for linking files")],
                             on_conflict: Annotated[ON_CONFLICT_LOOSE, typer.Option(..., "--on-conflict", "-o", help="Action to take on conflict")] = "throw-error",
                             which: Annotated[Optional[str], typer.Option(..., "--which", "-w", help="Specific items to process")] = "all",
                             interactive: Annotated[bool, typer.Option(..., "--interactive", "-i", help="Run in interactive mode")] = False):
    from machineconfig.profile.create_links import ConfigMapper, read_mapper
    mapper_full = read_mapper()["private"]
    if which is None:
        if interactive is False:
            typer.echo("[red]Error:[/] --which must be provided when not running in interactive mode.")
            typer.Exit(code=1)
            return
        from machineconfig.utils.options import choose_from_options
        items_chosen = choose_from_options(msg="Which symlink to create?", options=list(mapper_full.keys()), tv=True, multi=True)
    else:
        if interactive is True:
            typer.echo("[yellow]Warning:[/] --which is provided, but its not allowed to be used together with --interactive. Ignoring --interactive flag.")
            typer.Exit(code=0)
            return
        if which == "all":
            items_chosen = list(mapper_full.keys())
        else:
            items_chosen = which.split(",")
    items_objections: dict[str, list[ConfigMapper]] = {item: mapper_full[item] for item in items_chosen if item in mapper_full}
    if len(items_objections) == 0:
        typer.echo("[red]Error:[/] No valid items selected.")
        typer.Exit(code=1)
        return
    from machineconfig.profile.create_links import apply_mapper
    method_map: dict[str, Literal["symlink", "copy"]] = {
        "s": "symlink",
        "symlink": "symlink",
        "c": "copy",
        "copy": "copy",
    }
    method = method_map[method]
    apply_mapper(mapper_data=items_objections, on_conflict=ON_CONFLICT_MAPPER[on_conflict], method=method)



from machineconfig.jobs.installer.package_groups import PACKAGE_GROUP2NAMES
from machineconfig.utils.schemas.installer.installer_types import InstallerData
from pathlib import Path
from machineconfig.utils.path_extended import DECOMPRESS_SUPPORTED_FORMATS, PathExtended
from machineconfig.utils.source_of_truth import INSTALL_TMP_DIR


def get_group_name_to_repr() -> dict[str, str]:
    # Build category options and maintain a mapping from display text to actual category name
    category_display_to_name: dict[str, str] = {}
    for group_name, group_values in PACKAGE_GROUP2NAMES.items():
        display = f"📦 {group_name:<20}" + "   --   " + f"{'|'.join(group_values):<60}"
        category_display_to_name[display] = group_name
    return category_display_to_name


def handle_installer_not_found(search_term: str, app_apps: list[InstallerData]) -> None:  # type: ignore
    """Handle installer not found with friendly suggestions using fuzzy matching."""
    from difflib import get_close_matches
    from rich.console import Console
    from rich.panel import Panel
    from rich.table import Table
    all_names = sorted([inst["appName"] for inst in app_apps])
    name_to_doc = {inst["appName"]: inst["doc"] for inst in app_apps}
    all_descriptions = {f"{inst['appName']}: {inst['doc']}": inst["appName"] for inst in app_apps}

    close_name_matches = get_close_matches(search_term, all_names, n=5, cutoff=0.4)
    close_description_matches = get_close_matches(search_term, list(all_descriptions.keys()), n=5, cutoff=0.4)

    search_lower = search_term.lower()
    substring_matches = [
        inst["appName"]
        for inst in app_apps
        if search_lower in inst["appName"].lower() or search_lower in inst["doc"].lower()
    ]

    ordered_matches: list[str] = list(
        dict.fromkeys(
            close_name_matches
            + [all_descriptions[desc] for desc in close_description_matches]
            + substring_matches
        )
    )
    top_matches = ordered_matches[:10]
    console = Console()

    console.print(f"\n❌ '[red]{search_term}[/red]' was not found.", style="bold")
    if top_matches:
        console.print("🤔 Did you mean one of these?", style="yellow")
        table = Table(show_header=True, header_style="bold", box=None, pad_edge=False)
        table.add_column("#", justify="right", width=3)
        table.add_column("Installer", style="green")
        table.add_column("Description", style="dim", overflow="fold")
        for i, match in enumerate(top_matches, 1):
            table.add_row(f"[cyan]{i}[/cyan]", match, name_to_doc.get(match, ""))
        console.print(table)
    else:
        console.print("📋 Here are some available options:", style="blue")
        # Show first 10 installers as examples
        if len(all_names) > 10:
            sample_names = all_names[:10]
        else:
            sample_names = all_names
        table = Table(show_header=True, header_style="bold", box=None, pad_edge=False)
        table.add_column("#", justify="right", width=3)
        table.add_column("Installer", style="green")
        table.add_column("Description", style="dim", overflow="fold")
        for i, name in enumerate(sample_names, 1):
            table.add_row(f"[cyan]{i}[/cyan]", name, name_to_doc.get(name, ""))
        console.print(table)
        if len(all_names) > 10:
            console.print(f"   [dim]... and {len(all_names) - 10} more[/dim]")

    panel = Panel(f"[bold blue]💡 Use 'ia' to interactively browse all available installers.[/bold blue]\n[bold blue]💡 Use one of the categories: {list(PACKAGE_GROUP2NAMES.keys())}[/bold blue]", title="[yellow]Helpful Tips[/yellow]", border_style="yellow")
    console.print(panel)


def install_deb_package(downloaded: Path) -> None:
    from rich import print as rprint
    from rich.panel import Panel
    print(f"📦 Installing .deb package: {downloaded}")
    import platform
    import subprocess
    assert platform.system() == "Linux"
    result = subprocess.run(f"sudo nala install -y {downloaded}", shell=True, capture_output=True, text=True)
    success = result.returncode == 0 and result.stderr == ""
    if not success:
        from rich.console import Group
        desc = "Installing .deb"
        sub_panels = []
        if result.stdout:
            sub_panels.append(Panel(result.stdout, title="STDOUT", style="blue"))
        if result.stderr:
            sub_panels.append(Panel(result.stderr, title="STDERR", style="red"))
        group_content = Group(f"❌ {desc} failed\nReturn code: {result.returncode}", *sub_panels)
        rprint(Panel(group_content, title=desc, style="red"))
    print("🗑️  Cleaning up .deb package...")
    if downloaded.is_file():
        downloaded.unlink(missing_ok=True)
    elif downloaded.is_dir():
        import shutil
        shutil.rmtree(downloaded, ignore_errors=True)


def download_and_prepare(download_url: str) -> PathExtended:
    # archive_path = PathExtended(download_url).download(folder=INSTALL_TMP_DIR)
    from machineconfig.scripts.python.helpers.helpers_utils.download import download
    downloaded_object = download(download_url, output_dir=str(INSTALL_TMP_DIR))
    if downloaded_object is None:
        raise ValueError(f"Failed to download from URL: {download_url}")
    archive_path = PathExtended(downloaded_object)
    extracted_path = archive_path
    if extracted_path.is_file() and any(ext in archive_path.suffixes for ext in DECOMPRESS_SUPPORTED_FORMATS):
        extracted_path = archive_path.decompress()
        # print(f"Decompressed {archive_path} to {extracted_path}")
        archive_path.delete(sure=True)
        if extracted_path.is_dir():
            nested_items = list(extracted_path.glob("*"))
            if len(nested_items) == 1:
                nested_path = PathExtended(nested_items[0])
                if nested_path.is_file() and any(ex in nested_path.suffixes for ex in DECOMPRESS_SUPPORTED_FORMATS):
                    extracted_path = nested_path.decompress()
                    nested_path.delete(sure=True)
    elif extracted_path.is_dir() and len([p for p in extracted_path.rglob("*") if not p.name.startswith(".")]) == 1:
        only_file_in = next(extracted_path.glob("*"))
        if only_file_in.is_file() and any(ext in str(only_file_in) for ext in DECOMPRESS_SUPPORTED_FORMATS):  # further decompress
            extracted_path = only_file_in.decompress()
            only_file_in.delete(sure=True)
    return extracted_path




from pathlib import Path

UV = Path(__file__).parent.joinpath("uv.ps1")

INTERACTIVE = Path(__file__).parent.joinpath("web_shortcuts/interactive.ps1")
LIVE =        Path(__file__).parent.joinpath("web_shortcuts/live_from_github.ps1")




"""
Installation Utilities
======================

This module provides functionality to download and install pre-checked applications.
"""

import csv
import platform
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Optional

import gdown
from rich.console import Console

from machineconfig.utils.path_extended import PathExtended
from machineconfig.utils.source_of_truth import CONFIG_ROOT

# Constants
APP_SUMMARY_PATH = CONFIG_ROOT.joinpath(f"profile/records/{platform.system().lower()}/apps_summary_report.csv")
CLOUD_STORAGE_NAME = "gdw"  # Default cloud storage name for rclone

console = Console()

def upload_app(path: PathExtended) -> Optional[str]:
    """Uploads the app to cloud storage and returns the shareable link."""
    try:
        # Using PathExtended.to_cloud
        # Assuming 'gdw' is the configured remote in rclone
        link_path = path.to_cloud(CLOUD_STORAGE_NAME, rel2home=True, share=True, os_specific=True, verbose=False)
        if link_path:
             return str(link_path)
        return None
    except Exception as e:
        console.print(f"[red]Failed to upload {path}: {e}[/red]")
        return None

def download_google_drive_file(url: str) -> PathExtended:
    """Downloads a file from Google Drive using gdown."""
    try:
        # Extract ID from URL
        # Assuming URL format like https://drive.google.com/file/d/FILE_ID/view or similar
        # or https://drive.google.com/uc?id=FILE_ID
        if "id=" in url:
            file_id = url.split("id=")[1].split("&")[0]
        elif "/d/" in url:
            file_id = url.split("/d/")[1].split("/")[0]
        else:
            # Fallback to gdown's auto detection or just pass URL
            file_id = url

        # Create a temporary directory for download
        output_dir = PathExtended.tmpdir(prefix="gdown_")
        # gdown.download returns the output filename
        output_file = gdown.download(id=file_id, output=str(output_dir) + "/", quiet=False, fuzzy=True)
        
        if not output_file:
            raise ValueError(f"Download failed for {url}")
            
        return PathExtended(output_file)
    except Exception as e:
        raise RuntimeError(f"Failed to download from Google Drive: {e}") from e

def install_cli_app(app_url: str) -> bool:
    """Downloads and installs a CLI app."""
    try:
        if not app_url:
            return False
            
        exe_path = download_google_drive_file(app_url)
        console.print(f"[green]Downloaded {exe_path.name}[/green]")
        
        system = platform.system().lower()
        if system in ["linux", "darwin"]:
            exe_path.chmod(0o755) # Make executable
            # Move to local bin (requires user to have write access or sudo, which we can't easily do here without interaction)
            # Using ~/.local/bin is safer
            install_path = Path.home() / ".local/bin"
            install_path.mkdir(parents=True, exist_ok=True)
            target = install_path / exe_path.name
            exe_path.rename(target)
            console.print(f"[green]Installed to {target}[/green]")
        elif system == "windows":
            install_path = Path.home() / "AppData/Local/Microsoft/WindowsApps"
            install_path.mkdir(parents=True, exist_ok=True)
            target = install_path / exe_path.name
            # Use shutil.move or PathExtended.move
            exe_path.move(path=target, overwrite=True)
            console.print(f"[green]Installed to {target}[/green]")
        
        return True
    except Exception as e:
        console.print(f"[red]Failed to install app from {app_url}: {e}[/red]")
        return False

def load_summary_report() -> list[dict[str, Any]]:
    """Loads the summary report from CSV."""
    if APP_SUMMARY_PATH.exists():
        with open(APP_SUMMARY_PATH, 'r', encoding='utf-8') as csvfile:
            reader = csv.DictReader(csvfile)
            return list(reader)
    else:
        console.print(f"[yellow]Warning: Summary report not found at {APP_SUMMARY_PATH}[/yellow]")
        return []

def download_safe_apps(name: str = "essentials") -> None:
    """Downloads and installs safe apps."""
    data = load_summary_report()
    if not data:
        console.print("[red]No app data available to install.[/red]")
        return

    apps_to_install = []
    if name == "essentials":
        # Install all available apps in the report? Or filter?
        # Original code installed all if name="essentials" (implied by logic)
        apps_to_install = [item['app_url'] for item in data if item.get('app_url')]
    else:
        apps_to_install = [item['app_url'] for item in data if item.get('app_name') == name and item.get('app_url')]

    if not apps_to_install:
        console.print(f"[yellow]No apps found to install for '{name}'.[/yellow]")
        return

    console.print(f"[bold]Installing {len(apps_to_install)} apps...[/bold]")
    
    with ThreadPoolExecutor(max_workers=5) as executor:
        results = list(executor.map(install_cli_app, apps_to_install))
    
    success_count = sum(results)
    console.print(f"[bold green]Successfully installed {success_count}/{len(apps_to_install)} apps.[/bold green]")



from pathlib import Path
from typing import Optional, Annotated
from machineconfig.scripts.python.helpers.helpers_devops.cli_share_file import share_file_receive, share_file_send
import typer


def display_share_url(local_ip_v4: str, port: int, protocol: str = "http") -> None:
    """Display a flashy, unmissable share URL announcement."""
    from rich.console import Console
    from rich.panel import Panel
    from rich.text import Text
    from rich.align import Align
    console = Console()
    # Create the main message with styling
    url_text = Text(f"{protocol}://{local_ip_v4}:{port}", style="bold bright_cyan underline")
    message = Text.assemble(
        ("🚀 ", "bright_red"),
        ("Share server is now accessible at: ", "bright_white bold"),
        url_text,
        (" 🚀", "bright_red")
    )
    # Create a fancy panel with borders and styling
    panel = Panel(
        Align.center(message),
        title="[bold bright_green]🌐 SHARE SERVER READY 🌐[/bold bright_green]",
        subtitle="[italic bright_yellow]⚡ Click the link above to access your shared files! ⚡[/italic bright_yellow]",
        border_style="bright_magenta",
        padding=(1, 2),
        expand=False
    )
    # Print with extra spacing and attention-grabbing elements
    console.print(panel)


def web_file_explorer(
    path: Annotated[str, typer.Argument(help="Path to the file or directory to share")],
    port: Annotated[Optional[int], typer.Option("--port", "-p", help="Port to run the share server on (default: 8080)")] = None,
    username: Annotated[Optional[str], typer.Option("--username", "-u", help="Username for share access (default: current user)")] = None,
    no_auth: Annotated[bool, typer.Option("--no-auth", "-na", help="Disable authentication for share access")] = False,
    password: Annotated[Optional[str], typer.Option("--password", "-w", help="Password for share access (default: from ~/dotfiles/creds/passwords/quick_password)")] = None,
    bind_address: Annotated[str, typer.Option("--bind", "-a", help="Address to bind the server to")] = "0.0.0.0",
    over_internet: Annotated[bool, typer.Option("--over-internet", "-i", help="Expose the share server over the internet using ngrok")] = False,
    backend: Annotated[str, typer.Option("--backend", "-b", help="Backend to use: filebrowser (default), miniserve, qrcp, or easy-sharing")] = "miniserve",
) -> None:
    from machineconfig.utils.installer_utils.installer_cli import install_if_missing

    if backend not in ["filebrowser", "miniserve", "qrcp", "easy-sharing"]:
        typer.echo(f"❌ ERROR: Invalid backend '{backend}'. Must be one of: filebrowser, miniserve, qrcp, easy-sharing", err=True)
        raise typer.Exit(code=1)
    install_if_missing(which=backend)
    if over_internet:
        install_if_missing(which="ngrok")
    if password is None and not no_auth:
        pwd_path = Path.home().joinpath("dotfiles/creds/passwords/quick_password")
        if pwd_path.exists():
            password = pwd_path.read_text(encoding="utf-8").strip()
            if username is None:
                import getpass
                username = getpass.getuser()
        # else:
        #     typer.echo(f"⚠️  WARNING: Password not provided and default password file does not exist.\nPath: {pwd_path}\nUsing default password: 'quick_password' (insecure!)", err=True)
        #     raise typer.Exit(code=1)


    if port is None:
        port = 8080

    import machineconfig.scripts.python.helpers.helpers_network.address as helper
    res = helper.select_lan_ipv4(prefer_vpn=False)
    if res is None:
        typer.echo("❌ ERROR: Could not determine local LAN IPv4 address for share server.", err=True)
        raise typer.Exit(code=1)
    local_ip_v4 = res

    protocol = "http"
    display_share_url(local_ip_v4, port, protocol)
    
    path_obj = Path(path).resolve()
    if not path_obj.exists():
        typer.echo(f"❌ ERROR: Path does not exist: {path}", err=True)
        raise typer.Exit(code=1)
    
    if backend == "filebrowser":
        db_path = Path.home().joinpath(".config/filebrowser/filebrowser.db")
        db_path.parent.mkdir(parents=True, exist_ok=True)
        command = f"""
filebrowser users add {username} "{password}" --database {db_path}
filebrowser --address {bind_address} --port {port} --root "{path_obj}" --database {db_path}
"""
    elif backend == "miniserve":
        if username and password:
            auth_line = f"""--auth "{username}:{password}" """
        else:
            auth_line = ""
        command = f"""miniserve --port {port} --interfaces {bind_address} {auth_line} --upload-files --mkdir --enable-tar --enable-tar-gz --enable-zip --qrcode "{path_obj}" """
    elif backend == "easy-sharing":
        if username is None and password is None:
            auth_line = ""
        else:
            auth_line = f'--username "{username}" --password "{password}"'
        command = f"""easy-sharing --port {port} {auth_line} "{path_obj}" """
    elif backend == "qrcp":
        command = f"""qrcp --port {port} --bind {bind_address} "{path_obj}" """
    else:
        typer.echo(f"❌ ERROR: Unknown backend '{backend}'", err=True)
        raise typer.Exit(code=1)

    from machineconfig.utils.code import exit_then_run_shell_script
    exit_then_run_shell_script(script=command, strict=False)
    # import subprocess
    # import time
    # server_process: subprocess.Popen[bytes]
    # server_process = subprocess.Popen(command, shell=True)
    # processes = [server_process]
    # if over_internet:
    #     ngrok_process = subprocess.Popen(f"ngrok http {port}", shell=True)
    #     processes.append(ngrok_process)
    #     time.sleep(3)
    #     try:
    #         import requests
    #         response = requests.get("http://localhost:4040/api/tunnels")
    #         data = response.json()
    #         public_url = data['tunnels'][0]['public_url']
    #         print(f"🌐 Ngrok tunnel ready: {public_url}")
    #     except Exception as e:
    #         print(f"Could not retrieve ngrok URL: {e}")
    
    # try:
    #     while True:
    #         print(f"Share server ({backend}) is running. Press Ctrl+C to stop.")
    #         time.sleep(2)
    # except KeyboardInterrupt:
    #     print("\nTerminating processes...")
    #     for p in processes:
    #         p.terminate()
    #         p.wait()


def get_share_file_app():
    app = typer.Typer(name="share-file", help="Send or receive files using croc with relay server.")
    app.command(name="send", no_args_is_help=True, hidden=False, help="[s] send files from here.")(share_file_send)
    app.command(name="s", no_args_is_help=True, hidden=True, help="[s] send files from here.")(share_file_send)
    app.command(name="receive", no_args_is_help=True, hidden=False, help="[r] receive files to here.")(share_file_receive)
    app.command(name="r", no_args_is_help=True, hidden=True, help="[r] receive files to here.")(share_file_receive)
    return app

def main_with_parser():
    typer.run(web_file_explorer)


if __name__ == "__main__":
    pass



from typing import Optional
import platform
from machineconfig.utils.schemas.installer.installer_types import InstallerData


def main(installer_data: InstallerData, version: Optional[str]):
    _ = installer_data, version
    system = platform.system()
    if system == "Windows":
        raise NotImplementedError("Installer is not yet implemented for Windows.")
    elif system == "Linux":
        from pathlib import Path
        import machineconfig.jobs.installer as module
        program = Path(module.__file__).parent.joinpath("linux_scripts/cloudflare_warp_cli.sh").read_text(encoding="utf-8")
    elif system == "Darwin":
        program = "brew install --cask cloudflare-warp"
    else:
        raise NotImplementedError(f"Unsupported platform: {system}")
    import subprocess
    subprocess.run(program, shell=True, check=True)
    return f"Cloudflare WARP CLI installed successfully on {system}."






#!/usr/bin/env python3
import base64
import pathlib
import subprocess
import tempfile
import os


def main(options_to_preview_mapping: dict[str, str]) -> str | None:
    keys = list(options_to_preview_mapping.keys())
    if not keys:
        return None
    with tempfile.TemporaryDirectory(prefix="tv_channel_") as tmpdir:
        tempdir = pathlib.Path(tmpdir)
        entries: list[str] = []
        index_map: dict[int, str] = {}
        preview_map_path = tempdir / "previews.tsv"
        preview_rows: list[str] = []
        for idx, key in enumerate(keys):
            display_key = key.replace("\t", "    ").replace("\n", " ")
            entries.append(f"{idx}\t{display_key}")
            index_map[idx] = key
            encoded_preview = base64.b64encode(options_to_preview_mapping[key].encode("utf-8")).decode("ascii")
            preview_rows.append(f"{idx}\t{encoded_preview}")
        preview_map_path.write_text("\n".join(preview_rows), encoding="utf-8")
        entries_path = tempdir / "entries.tsv"
        entries_path.write_text("\n".join(entries), encoding="utf-8")
        preview_script = tempdir / "preview.sh"
        preview_script.write_text(
            """#!/usr/bin/env bash
set -euo pipefail

idx="$1"
script_dir="$(cd -- "$(dirname -- "$0")" && pwd)"
previews_file="${script_dir}/previews.tsv"

if [[ ! -f "${previews_file}" ]]; then
    exit 0
fi

encoded_preview="$(awk -F '\t' -v idx="${idx}" '($1==idx){print $2; exit}' "${previews_file}" || true)"

if [[ -z "${encoded_preview}" ]]; then
    exit 0
fi

preview_content="$(printf '%s' "${encoded_preview}" | base64 --decode)"

if command -v bat >/dev/null 2>&1; then
    printf '%s' "${preview_content}" | glow -
elif command -v bat >/dev/null 2>&1; then
    printf '%s' "${preview_content}" | bat --language=markdown --color=always --style=plain --paging=never
elif command -v glow >/dev/null 2>&1; then
    printf '%s' "${preview_content}" | glow -
else
    printf '%s' "${preview_content}"
fi
""",
            encoding="utf-8"
        )
        preview_script.chmod(0o755)
        channel_config = f"""[metadata]
name = "temp_options"
description = "Temporary channel for selecting options"

[source]
command = "cat '{entries_path}'"
display = "{{split:\\t:1}}"
output = "{{split:\\t:0}}"

[preview]
command = "{preview_script} {{split:\\t:0}}"

[ui.preview_panel]
size = 50
"""
        channel_path = tempdir / "temp_options.toml"
        channel_path.write_text(channel_config, encoding="utf-8")
        env = os.environ.copy()
        tv_config_dir = pathlib.Path.home() / ".config" / "television"
        if not tv_config_dir.exists():
            tv_config_dir = pathlib.Path(os.getenv("XDG_CONFIG_HOME", str(pathlib.Path.home() / ".config"))) / "television"
        cable_dir = tv_config_dir / "cable"
        cable_dir.mkdir(parents=True, exist_ok=True)
        temp_channel_link = cable_dir / "temp_options.toml"
        if temp_channel_link.exists() or temp_channel_link.is_symlink():
            temp_channel_link.unlink()
        temp_channel_link.symlink_to(channel_path)
        output_file = tempdir / "selection.txt"
        try:
            result = subprocess.run(["tv", "temp_options"], check=False, stdout=output_file.open("w"), text=True, env=env)
        finally:
            if temp_channel_link.exists() or temp_channel_link.is_symlink():
                temp_channel_link.unlink()
        if result.returncode not in (0, 130):
            raise SystemExit(result.returncode)
        if result.returncode == 130:
            return None
        if not output_file.exists():
            return None
        selected = output_file.read_text().strip()
        if not selected:
            return None
        try:
            index = int(selected)
        except ValueError:
            return None
        return index_map.get(index)


if __name__ == "__main__":
    demo_mapping = {
        "Option 1": "# Option 1\nThis is the preview for option 1.",
        "Option 2": "# Option 2\nThis is the preview for option 2.",
        "Option 3": "# Option 3\nThis is the preview for option 3."
    }
    main(demo_mapping)



#!/usr/bin/env -S uv run --no-dev --project

"""croshell - Cross-shell command execution."""

from typing import Annotated, Optional
import typer


def croshell(
    path: Annotated[Optional[str], typer.Argument(help="path of file to read.")] = None,
    project_path: Annotated[Optional[str], typer.Option("--project", "-p", help="specify uv project to use")] = None,
    uv_with: Annotated[Optional[str], typer.Option("--uv-with", "-w", help="specify uv with packages to use")] = None,
    marimo: Annotated[bool, typer.Option("--marimo", "-m", help="open the notebook using marimo if available")] = False,
    jupyter: Annotated[bool, typer.Option("--jupyter", "-j", help="run in jupyter interactive console")] = False,
    vscode: Annotated[bool, typer.Option("--vscode", "-c", help="open the script in vscode")] = False,
    visidata: Annotated[bool, typer.Option("--visidata", "-v", help="open data file in visidata")] = False,
    python: Annotated[bool, typer.Option("--python", "-P", help="flag to use python over IPython.")] = False,
    profile: Annotated[Optional[str], typer.Option("--profile", "-r", help="ipython profile to use, defaults to default profile.")] = None,
) -> None:
    """Cross-shell command execution."""
    from machineconfig.scripts.python.helpers.helpers_croshell.croshell_impl import croshell as impl
    impl(path=path, project_path=project_path, uv_with=uv_with, marimo=marimo, jupyter=jupyter, vscode=vscode, visidata=visidata, python=python, profile=profile)


def main() -> None:
    typer.run(croshell)


if __name__ == "__main__":
    main()





from platform import system
from pathlib import Path
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich import box
import subprocess
import os
import re


console = Console()


def _run(cmd: list[str]) -> tuple[bool, str]:
    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=False)
        return result.returncode == 0, result.stdout.strip()
    except FileNotFoundError:
        return False, ""


def _check_sshd_installed() -> tuple[bool, str]:
    sshd_paths = ["/usr/sbin/sshd", "/usr/bin/sshd", "/sbin/sshd"]
    for path in sshd_paths:
        if Path(path).exists():
            return True, path
    ok, which_out = _run(["which", "sshd"])
    if ok and which_out:
        return True, which_out
    return False, ""


def _detect_package_manager() -> tuple[str, str]:
    if Path("/usr/bin/apt").exists() or Path("/usr/bin/apt-get").exists():
        return "apt", "sudo apt update && sudo apt install -y openssh-server"
    if Path("/usr/bin/dnf").exists():
        return "dnf", "sudo dnf install -y openssh-server"
    if Path("/usr/bin/yum").exists():
        return "yum", "sudo yum install -y openssh-server"
    if Path("/usr/bin/pacman").exists():
        return "pacman", "sudo pacman -S --noconfirm openssh"
    if Path("/usr/bin/zypper").exists():
        return "zypper", "sudo zypper install -y openssh"
    return "unknown", "# Install openssh-server using your package manager"


def ssh_debug_linux() -> dict[str, dict[str, str | bool]]:
    if system() != "Linux":
        raise NotImplementedError("ssh_debug_linux is only supported on Linux")

    results: dict[str, dict[str, str | bool]] = {}
    issues: list[tuple[str, str, str]] = []
    current_user = os.environ.get("USER", os.environ.get("USERNAME", "unknown"))
    ssh_port = "22"
    ip_addresses: list[str] = []

    ok, hostname = _run(["hostname"])
    hostname = hostname if ok else "unknown"

    install_info: list[str] = []
    sshd_installed, sshd_path = _check_sshd_installed()
    _pkg_manager, install_cmd = _detect_package_manager()
    if not sshd_installed:
        results["installation"] = {"status": "error", "message": "OpenSSH Server not installed"}
        issues.append(("sshd not installed", "Cannot accept incoming SSH connections", install_cmd))
        install_info.append("❌ OpenSSH Server: [red]NOT INSTALLED[/red]")
        install_info.append(f"   [dim]Install with: {install_cmd}[/dim]")
    else:
        results["installation"] = {"status": "ok", "message": f"sshd found at {sshd_path}"}
        install_info.append(f"✅ OpenSSH Server: installed at [cyan]{sshd_path}[/cyan]")
    console.print(Panel("\n".join(install_info), title="[bold]Installation[/bold]", border_style="blue"))

    ssh_dir = Path.home().joinpath(".ssh")
    authorized_keys = ssh_dir.joinpath("authorized_keys")
    home_dir = Path.home()

    perm_info: list[str] = []
    home_stat = os.stat(home_dir)
    home_perms = oct(home_stat.st_mode)[-3:]
    if home_perms[2] in ["7", "6", "3", "2"]:
        results["home_directory"] = {"status": "error", "message": f"Home world-writable: {home_perms}"}
        issues.append((f"Home dir perms {home_perms}", "sshd refuses login if home is world-writable", f"chmod 755 {home_dir}"))
        perm_info.append(f"❌ Home directory: [red]{home_perms}[/red] (world-writable)")
        perm_info.append("   [dim]sshd will refuse key auth if home is writable by others[/dim]")
    else:
        perm_info.append(f"✅ Home directory: {home_perms}")

    if not ssh_dir.exists():
        results["ssh_directory"] = {"status": "error", "message": "~/.ssh missing"}
        issues.append(("~/.ssh missing", "No place for authorized_keys", "mkdir -p ~/.ssh && chmod 700 ~/.ssh"))
        perm_info.append("❌ ~/.ssh: [red]does not exist[/red]")
    else:
        ssh_perms = oct(os.stat(ssh_dir).st_mode)[-3:]
        if ssh_perms != "700":
            results["ssh_directory"] = {"status": "error", "message": f"~/.ssh perms {ssh_perms}"}
            issues.append((f"~/.ssh perms {ssh_perms}", "sshd requires 700 on ~/.ssh", "chmod 700 ~/.ssh"))
            perm_info.append(f"❌ ~/.ssh: [red]{ssh_perms}[/red] (must be 700)")
        else:
            perm_info.append(f"✅ ~/.ssh: {ssh_perms}")

    if not authorized_keys.exists():
        results["authorized_keys"] = {"status": "error", "message": "authorized_keys missing"}
        issues.append(("authorized_keys missing", "No keys = no login", "Add public key: cat id_rsa.pub >> ~/.ssh/authorized_keys && chmod 600 ~/.ssh/authorized_keys"))
        perm_info.append("❌ authorized_keys: [red]does not exist[/red]")
        perm_info.append("   [dim]No authorized keys = cannot login with SSH key[/dim]")
    else:
        ak_perms = oct(os.stat(authorized_keys).st_mode)[-3:]
        try:
            keys = [line for line in authorized_keys.read_text(encoding="utf-8").split("\n") if line.strip()]
            key_count = len(keys)
        except Exception:
            key_count = 0
        if ak_perms not in ["600", "644"]:
            results["authorized_keys"] = {"status": "error", "message": f"authorized_keys perms {ak_perms}"}
            issues.append((f"authorized_keys perms {ak_perms}", "sshd requires 600 or 644", "chmod 600 ~/.ssh/authorized_keys"))
            perm_info.append(f"❌ authorized_keys: [red]{ak_perms}[/red] ({key_count} key(s)) - must be 600/644")
        else:
            results["authorized_keys"] = {"status": "ok", "message": f"{key_count} key(s)"}
            perm_info.append(f"✅ authorized_keys: {ak_perms} ([green]{key_count} key(s)[/green])")

    console.print(Panel("\n".join(perm_info), title="[bold]Permissions[/bold]", border_style="blue"))

    svc_info: list[str] = []
    ssh_ok, _ = _run(["systemctl", "is-active", "ssh"])
    sshd_ok, _ = _run(["systemctl", "is-active", "sshd"])
    if ssh_ok or sshd_ok:
        svc_name = "ssh" if ssh_ok else "sshd"
        results["ssh_service"] = {"status": "ok", "message": f"{svc_name} running"}
        svc_info.append(f"✅ Service: [green]{svc_name} running[/green]")
    else:
        results["ssh_service"] = {"status": "error", "message": "sshd not running"}
        issues.append(("sshd not running", "No SSH daemon = no connections", "sudo systemctl start ssh && sudo systemctl enable ssh"))
        svc_info.append("❌ Service: [red]not running[/red]")

    console.print(Panel("\n".join(svc_info), title="[bold]Service[/bold]", border_style="blue"))

    net_info: list[str] = []
    ok, ip_out = _run(["ip", "addr", "show"])
    if ok:
        ip_addresses = re.findall(r'inet\s+(\d+\.\d+\.\d+\.\d+)/\d+.*scope\s+global', ip_out)
        if ip_addresses:
            net_info.append(f"🌐 IP: [cyan]{', '.join(ip_addresses)}[/cyan]")

    sshd_config_paths = [Path("/etc/ssh/sshd_config"), Path("/etc/sshd_config")]
    sshd_config: Path | None = None
    for p in sshd_config_paths:
        if p.exists():
            sshd_config = p
            break

    if sshd_config:
        try:
            config_text = sshd_config.read_text(encoding="utf-8")
            port_lines = [line for line in config_text.split("\n") if line.strip().startswith("Port") and not line.strip().startswith("#")]
            if port_lines:
                ssh_port = port_lines[0].split()[1]
            net_info.append(f"🔌 Port: [cyan]{ssh_port}[/cyan]")

            pubkey_lines = [line for line in config_text.split("\n") if "PubkeyAuthentication" in line and not line.strip().startswith("#")]
            if pubkey_lines and "no" in pubkey_lines[-1].lower():
                results["pubkey_auth"] = {"status": "error", "message": "PubkeyAuthentication disabled"}
                issues.append(("PubkeyAuthentication disabled", "Key-based login won't work", f"Edit {sshd_config}: set PubkeyAuthentication yes, then sudo systemctl restart ssh"))
                net_info.append("❌ PubkeyAuthentication: [red]disabled[/red]")
            else:
                net_info.append("✅ PubkeyAuthentication: enabled")

            password_lines = [line for line in config_text.split("\n") if "PasswordAuthentication" in line and not line.strip().startswith("#")]
            if password_lines:
                password_enabled = "yes" in password_lines[-1].lower()
                if password_enabled:
                    results["password_auth"] = {"status": "ok", "message": "PasswordAuthentication enabled"}
                    net_info.append("✅ PasswordAuthentication: [green]enabled[/green]")
                else:
                    results["password_auth"] = {"status": "info", "message": "PasswordAuthentication disabled"}
                    net_info.append("ℹ️  PasswordAuthentication: [yellow]disabled[/yellow] (key-only)")
            else:
                results["password_auth"] = {"status": "ok", "message": "PasswordAuthentication enabled (default)"}
                net_info.append("✅ PasswordAuthentication: [green]enabled[/green] (default)")

            permit_root = [line for line in config_text.split("\n") if "PermitRootLogin" in line and not line.strip().startswith("#")]
            if permit_root:
                val = permit_root[-1].split()[-1].lower()
                net_info.append(f"ℹ️  PermitRootLogin: {val}")
        except Exception:
            pass

    ok, ss_out = _run(["ss", "-tlnp"])
    if ok:
        listening = [line for line in ss_out.split("\n") if f":{ssh_port}" in line]
        if not listening:
            results["ssh_listening"] = {"status": "error", "message": f"Not listening on {ssh_port}"}
            issues.append((f"Not listening on port {ssh_port}", "No connections possible", "sudo systemctl restart ssh"))
            net_info.append(f"❌ Listening: [red]NOT on port {ssh_port}[/red]")
        elif all("127.0.0.1" in line or "[::1]" in line for line in listening):
            results["ssh_listening"] = {"status": "error", "message": "Localhost only"}
            issues.append(("SSH bound to localhost", "Only local connections", f"Edit {sshd_config}: remove/comment ListenAddress 127.0.0.1"))
            net_info.append("❌ Listening: [red]localhost only[/red]")
        else:
            results["ssh_listening"] = {"status": "ok", "message": f"Listening on {ssh_port}"}
            net_info.append(f"✅ Listening: 0.0.0.0:{ssh_port}")

    fw_checked = False
    ok, ufw_out = _run(["ufw", "status"])
    if ok and "Status: active" in ufw_out:
        fw_checked = True
        if f"{ssh_port}/tcp" in ufw_out.lower() or "ssh" in ufw_out.lower() or f" {ssh_port} " in ufw_out:
            results["firewall"] = {"status": "ok", "message": "UFW allows SSH"}
            net_info.append("✅ Firewall (UFW): allows SSH")
        else:
            results["firewall"] = {"status": "error", "message": "UFW blocking SSH"}
            issues.append(("UFW blocking SSH", "Incoming connections dropped", f"sudo ufw allow {ssh_port}/tcp"))
            net_info.append("❌ Firewall (UFW): [red]blocking SSH[/red]")
            net_info.append("   [dim]Active firewall without SSH rule = blocked[/dim]")

    if not fw_checked:
        ok, fwd_out = _run(["firewall-cmd", "--state"])
        if ok and "running" in fwd_out.lower():
            fw_checked = True
            ok2, svc_out = _run(["firewall-cmd", "--list-services"])
            if ok2 and "ssh" in svc_out.lower():
                results["firewall"] = {"status": "ok", "message": "firewalld allows SSH"}
                net_info.append("✅ Firewall (firewalld): allows SSH")
            else:
                results["firewall"] = {"status": "error", "message": "firewalld blocking SSH"}
                issues.append(("firewalld blocking SSH", "Incoming connections dropped", "sudo firewall-cmd --permanent --add-service=ssh && sudo firewall-cmd --reload"))
                net_info.append("❌ Firewall (firewalld): [red]blocking SSH[/red]")

    if not fw_checked:
        ok, ipt_out = _run(["iptables", "-L", "INPUT", "-n"])
        if ok and ipt_out:
            has_drop_policy = "policy DROP" in ipt_out or "policy REJECT" in ipt_out
            has_ssh_allow = f"dpt:{ssh_port}" in ipt_out or "dpt:ssh" in ipt_out
            if has_drop_policy and not has_ssh_allow:
                results["firewall"] = {"status": "error", "message": "iptables blocking SSH"}
                issues.append(("iptables blocking SSH", "DROP/REJECT policy without SSH allow", f"sudo iptables -I INPUT -p tcp --dport {ssh_port} -j ACCEPT"))
                net_info.append("❌ Firewall (iptables): [red]DROP policy, no SSH rule[/red]")
                fw_checked = True
            elif has_drop_policy and has_ssh_allow:
                results["firewall"] = {"status": "ok", "message": "iptables allows SSH"}
                net_info.append("✅ Firewall (iptables): allows SSH")
                fw_checked = True

    if not fw_checked:
        net_info.append("ℹ️  Firewall: none detected / not active")

    console.print(Panel("\n".join(net_info), title="[bold]Network & Firewall[/bold]", border_style="blue"))

    other_info: list[str] = []
    hosts_deny = Path("/etc/hosts.deny")
    if hosts_deny.exists():
        try:
            content = hosts_deny.read_text(encoding="utf-8")
            active = [line for line in content.splitlines() if line.strip() and not line.strip().startswith("#")]
            joined = " ".join(active).lower()
            if "sshd" in joined or "all" in joined:
                results["hosts_deny"] = {"status": "error", "message": "hosts.deny blocking"}
                issues.append(("hosts.deny blocking SSH", "TCP wrappers deny before firewall", "Edit /etc/hosts.deny to remove sshd/ALL entries"))
                other_info.append("❌ /etc/hosts.deny: [red]may block SSH[/red]")
            else:
                other_info.append("✅ /etc/hosts.deny: OK")
        except Exception:
            pass

    ok, se_out = _run(["getenforce"])
    if ok and se_out:
        if se_out == "Enforcing":
            other_info.append("ℹ️  SELinux: Enforcing (run [cyan]restorecon -Rv ~/.ssh[/cyan] if issues)")
        else:
            other_info.append(f"ℹ️  SELinux: {se_out}")

    log_files = [Path("/var/log/auth.log"), Path("/var/log/secure")]
    for lf in log_files:
        if lf.exists():
            ok, tail = _run(["tail", "-n", "20", str(lf)])
            if ok:
                errors = [line for line in tail.split("\n") if any(k in line.lower() for k in ["error", "failed", "refused", "denied"]) and "ssh" in line.lower()]
                if errors:
                    other_info.append(f"⚠️  Recent SSH errors in {lf.name}: {len(errors)}")
                else:
                    other_info.append(f"✅ {lf.name}: no recent SSH errors")
            break

    if other_info:
        console.print(Panel("\n".join(other_info), title="[bold]Additional[/bold]", border_style="blue"))

    if issues:
        fix_table = Table(title="Issues & Fixes", box=box.ROUNDED, show_lines=True, title_style="bold red")
        fix_table.add_column("Issue", style="yellow", width=25)
        fix_table.add_column("Impact", style="white", width=35)
        fix_table.add_column("Fix Command", style="green", width=55)
        for issue, impact, fix in issues:
            fix_table.add_row(issue, impact, fix)
        console.print(fix_table)

        fix_script_path = Path("/tmp/ssh_fix.sh")
        script_lines = ["#!/bin/bash", "set -e", "", "# SSH Fix Script - Generated by ssh_debug_linux", f"# {len(issues)} issue(s) to fix", ""]
        for issue, _impact, fix in issues:
            script_lines.append(f"# Fix: {issue}")
            script_lines.append(fix)
            script_lines.append("")
        script_lines.append("echo 'All fixes applied. Re-run ssh_debug_linux to verify.'")
        fix_script_path.write_text("\n".join(script_lines), encoding="utf-8")
        fix_script_path.chmod(0o755)

        console.print(Panel(f"[bold yellow]⚠️  {len(issues)} issue(s) found[/bold yellow]\n\nFix script generated: [cyan]{fix_script_path}[/cyan]\nRun: [green]sudo bash {fix_script_path}[/green]", title="[bold]Summary[/bold]", border_style="yellow"))
    else:
        conn_info = f"👤 {current_user}  🖥️  {hostname}  🔌 :{ssh_port}"
        if ip_addresses:
            conn_info += f"\n\n[bold]Connect:[/bold] ssh {current_user}@{ip_addresses[0]}"
        console.print(Panel(f"[bold green]✅ All checks passed[/bold green]\n\n{conn_info}", title="[bold]Ready[/bold]", border_style="green"))

    return results


if __name__ == "__main__":
    ssh_debug_linux()



import platform
import re
import shutil
import subprocess
from pathlib import Path

from machineconfig.utils.ssh_utils.wsl_helper import (
    ensure_relative_path,
    remove_path,
    ensure_wsl_environment,
    ensure_windows_environment,
    ensure_linux_environment,
    resolve_windows_home_from_wsl,
    resolve_wsl_home_on_windows,
    run_windows_copy_command,
    ensure_symlink,
    normalize_port_spec_for_firewall,
)


def copy_when_inside_wsl(source: Path | str, target: Path | str, overwrite: bool, windows_username: str | None) -> None:
    ensure_wsl_environment()
    source_relative = ensure_relative_path(source)
    target_relative = ensure_relative_path(target)
    source_path = Path.home() / source_relative
    target_path = resolve_windows_home_from_wsl(windows_username) / target_relative
    if not source_path.exists():
        raise FileNotFoundError(source_path)
    if target_path.exists():
        if not overwrite:
            raise FileExistsError(target_path)
        remove_path(target_path)
    target_path.parent.mkdir(parents=True, exist_ok=True)
    if source_path.is_dir():
        shutil.copytree(source_path, target_path, dirs_exist_ok=False)
        return
    print(f"Copying {source_path} to {target_path}")
    shutil.copy2(source_path, target_path)


def copy_when_inside_windows(source: Path | str, target: Path | str, overwrite: bool) -> None:
    ensure_windows_environment()
    source_relative = ensure_relative_path(source)
    target_relative = ensure_relative_path(target)
    source_path = Path.home() / source_relative
    target_path = resolve_wsl_home_on_windows() / target_relative
    if not source_path.exists():
        raise FileNotFoundError(source_path)
    if target_path.exists():
        if not overwrite:
            raise FileExistsError(target_path)
        remove_path(target_path)
    run_windows_copy_command(source_path, target_path)


def link_wsl_and_windows(windows_username: str | None) -> None:
    system = platform.system()
    if system == "Darwin":
        raise RuntimeError("link_wsl_and_windows is not designed for macOS")
    try:
        ensure_wsl_environment()
    except RuntimeError:
        try:
            ensure_windows_environment()
        except RuntimeError as exc:
            raise RuntimeError("link_wsl_and_windows must run inside Windows or WSL") from exc
        print("🔗 Running inside Windows, linking to WSL home...")
        target_path = resolve_wsl_home_on_windows()
        link_path = Path.home() / "wsl"
        created = ensure_symlink(link_path, target_path)
        if created:
            print(f"✅ Created symlink: {link_path} -> {target_path}")
        else:
            print(f"✅ Symlink already exists: {link_path} -> {target_path}")
        return
    print("🔗 Running inside WSL, linking to Windows home...")
    target_path = resolve_windows_home_from_wsl(windows_username)
    link_path = Path.home() / "win"
    created = ensure_symlink(link_path, target_path)
    if created:
        print(f"✅ Created symlink: {link_path} -> {target_path}")
    else:
        print(f"✅ Symlink already exists: {link_path} -> {target_path}")


def open_wsl_port(ports_spec: str) -> None:
    ensure_windows_environment()
    normalized_ports, description = normalize_port_spec_for_firewall(ports_spec)
    rule_name = f"WSL Ports {description}"
    # Build PowerShell array syntax for -LocalPort parameter (e.g., @('3000-4000','8080'))
    port_parts = normalized_ports.split(",")
    ps_array = "@(" + ",".join(f"'{p}'" for p in port_parts) + ")"
    script = f"New-NetFirewallRule -DisplayName '{rule_name}' -Direction Inbound -LocalPort {ps_array} -Protocol TCP -Action Allow"
    print(f"🔥 Opening firewall for ports: {description}...")
    result = subprocess.run(["powershell.exe", "-NoLogo", "-NoProfile", "-Command", script], capture_output=True, text=True)
    if result.returncode == 0:
        print(f"✅ Firewall rule created for ports: {description}")
    else:
        print(f"❌ Failed to create firewall rule: {result.stderr.strip()}")


def change_ssh_port(port: int) -> None:
    ensure_linux_environment()
    if port < 1 or port > 65535:
        raise ValueError(f"Invalid port number: {port}")

    sshd_config = Path("/etc/ssh/sshd_config")
    if not sshd_config.exists():
        raise FileNotFoundError(f"SSH config file not found: {sshd_config}")

    print(f"🔧 Changing SSH port to {port}...")

    content = sshd_config.read_text()
    new_content = re.sub(r"^#?\s*Port\s+\d+", f"Port {port}", content, flags=re.MULTILINE)
    if f"Port {port}" not in new_content:
        new_content = f"Port {port}\n" + new_content

    print(f"📝 Updating {sshd_config}...")
    result = subprocess.run(["sudo", "tee", str(sshd_config)], input=new_content.encode(), capture_output=True)
    if result.returncode != 0:
        raise RuntimeError(f"Failed to update sshd_config: {result.stderr.decode()}")
    print(f"✅ Updated {sshd_config}")

    override_dir = Path("/etc/systemd/system/ssh.socket.d")
    override_file = override_dir / "override.conf"
    override_content = f"""[Socket]
ListenStream=
ListenStream={port}
"""
    print(f"📝 Creating systemd socket override at {override_file}...")
    subprocess.run(["sudo", "mkdir", "-p", str(override_dir)], check=True)
    result = subprocess.run(["sudo", "tee", str(override_file)], input=override_content.encode(), capture_output=True)
    if result.returncode != 0:
        raise RuntimeError(f"Failed to create override file: {result.stderr.decode()}")
    print("✅ Created systemd socket override")

    print("🔄 Restarting SSH services...")
    subprocess.run(["sudo", "systemctl", "daemon-reload"], check=True)
    subprocess.run(["sudo", "systemctl", "restart", "ssh.socket"], check=False)
    subprocess.run(["sudo", "service", "ssh", "restart"], check=False)
    print(f"✅ SSH port changed to {port}")
    print(f"⚠️  Remember to open port {port} in Windows Firewall if running in WSL")


if __name__ == "__main__":
    copy_when_inside_wsl(Path("projects/source.txt"), Path("windows_projects/source.txt"), overwrite=True, windows_username=None)
    copy_when_inside_windows(Path("documents/example.txt"), Path("linux_documents/example.txt"), overwrite=True)



"""Pure Python implementation for msearch command - no typer dependencies."""


def machineconfig_search(
    directory: str,
    ast: bool,
    symantic: bool,
    extension: str,
    file: bool,
    no_dotfiles: bool,
    rga: bool,
    install_dependencies: bool,
) -> None:
    """Machineconfig search helper."""
    if install_dependencies:
        _install_dependencies()
        return
    if symantic:
        _run_symantic_search(extension=extension)
        return
    if ast:
        _run_ast_search(directory=directory)
        return
    if file:
        _run_file_search(no_dotfiles=no_dotfiles)
        return
    _run_text_search(rga=rga)


def _install_dependencies() -> None:
    """Install required dependencies."""
    from machineconfig.utils.installer_utils.installer_cli import install_if_missing
    install_if_missing("fzf")
    install_if_missing("tv")
    install_if_missing("bat")
    install_if_missing("fd")
    install_if_missing("rg")
    install_if_missing("rga")


def _run_symantic_search(extension: str) -> None:
    """Run symantic search."""
    script = ""
    for an_ex in extension.split(","):
        script = script + f"""\nparse *.{an_ex} """
    from machineconfig.utils.code import run_shell_script
    run_shell_script(script=script)


def _run_ast_search(directory: str) -> None:
    """Run AST search."""
    from machineconfig.scripts.python.helpers.helpers_search.ast_search import get_repo_symbols
    symbols = get_repo_symbols(directory)
    from machineconfig.utils.options import choose_from_options
    try:
        res = choose_from_options(options=symbols, msg="Select a symbol to search for:", tv=True, multi=False)
        from rich import print_json
        import json
        res_json = json.dumps(res, indent=4)
        print_json(res_json)
    except Exception as e:
        print(f"❌ Error during selection: {e}")


def _run_file_search(no_dotfiles: bool) -> None:
    """Run file search."""
    script = """fzf --ansi --preview-window 'right:60%' --preview 'bat --color=always --style=numbers,grid,header --line-range :300 {}' """
    if no_dotfiles:
        script = "fd | " + script
    from machineconfig.utils.code import run_shell_script
    run_shell_script(script=script)


def _run_text_search(rga: bool) -> None:
    """Run text search using fzf with ripgrep."""
    from machineconfig.scripts.python.helpers.helpers_msearch import FZFG_LINUX_PATH, FZFG_WINDOWS_PATH
    import platform
    if platform.system() == "Linux" or platform.system() == "Darwin":
        script = FZFG_LINUX_PATH.read_text(encoding="utf-8")
    elif platform.system() == "Windows":
        script = FZFG_WINDOWS_PATH.read_text(encoding="utf-8")
    else:
        raise RuntimeError("Unsupported platform")
    if rga:
        script = script.replace("rg ", "rga ").replace("ripgrep", "ripgrep-all")
    from machineconfig.utils.code import exit_then_run_shell_script
    exit_then_run_shell_script(script=script, strict=False)



"""Utility commands - lazy loading subcommands."""

import typer
from typing import Annotated, Optional, Literal


def kill_process(interactive: Annotated[bool, typer.Option(..., "--interactive", "-i", help="Interactively choose the process to kill")] = True) -> None:
    """⚔️ Choose a process to kill."""
    from machineconfig.utils.procs import main, ProcessManager
    if interactive:
        main()
        return
    _ = ProcessManager


def upgrade_packages() -> None:
    """⬆️ Upgrade project dependencies."""
    from machineconfig.utils.upgrade_packages import generate_uv_add_commands
    from pathlib import Path
    generate_uv_add_commands(pyproject_path=Path.cwd() / "pyproject.toml", output_path=Path.cwd() / "pyproject_init.sh")


def tui_env(which: Annotated[Literal["PATH", "p", "ENV", "e"], typer.Argument(help="Which environment variable to display.")] = "ENV") -> None:
    """📚 NAVIGATE ENV/PATH variable with TUI."""
    from machineconfig.scripts.python.helpers.helpers_utils.python import tui_env as impl
    impl(which=which)


def download(
    url: Annotated[Optional[str], typer.Argument(..., help="The URL to download the file from.")] = None,
    decompress: Annotated[bool, typer.Option(..., "--decompress", "-d", help="Decompress the file if it's an archive.")] = False,
    output: Annotated[Optional[str], typer.Option("--output", "-o", help="The output file path.")] = None,
    output_dir: Annotated[Optional[str], typer.Option("--output-dir", help="Directory to place the downloaded file in.")] = None,
) -> None:
    """⬇️ Download a file from a URL and optionally decompress it."""
    from machineconfig.scripts.python.helpers.helpers_utils.download import download as impl
    impl(url=url, decompress=decompress, output=output, output_dir=output_dir)


def get_machine_specs(hardware: Annotated[bool, typer.Option(..., "--hardware", "-h", help="Show compute capability")] = False) -> None:
    """💻 Get machine specifications."""
    from machineconfig.scripts.python.helpers.helpers_utils.python import get_machine_specs as impl
    impl(hardware=hardware)


def init_project(
    name: Annotated[Optional[str], typer.Option("--name", "-n", help="Name of the project.")] = None,
    tmp_dir: Annotated[bool, typer.Option("--tmp-dir", "-t", help="Use a temporary directory for the project initialization.")] = False,
    python: Annotated[Literal["3.11", "3.12", "3.13", "3.14"], typer.Option("--python", "-p", help="Python sub version for the uv virtual environment.")] = "3.13",
    libraries: Annotated[Optional[str], typer.Option("--libraries", "-l", help="Additional packages to include in the uv virtual environment (space separated).")] = None,
    group: Annotated[Optional[str], typer.Option("--group", "-g", help="group of packages names (no separation) p:plot, t:types, l:linting, i:interactive, d:data")] = "p,t,l,i,d",
) -> None:
    """🚀 Initialize a project with a uv virtual environment and install dev packages."""
    from machineconfig.scripts.python.helpers.helpers_utils.python import init_project as impl
    impl(name=name, tmp_dir=tmp_dir, python=python, libraries=libraries, group=group)


def edit_file_with_hx(path: Annotated[Optional[str], typer.Argument(..., help="The root directory of the project to edit, or a file path.")] = None) -> None:
    """✏️ Open a file in the default editor."""
    from machineconfig.scripts.python.helpers.helpers_utils.python import edit_file_with_hx as impl
    impl(path=path)


def merge_pdfs(
    pdfs: Annotated[list[str], typer.Argument(..., help="Paths to the PDF files to merge.")],
    output: Annotated[Optional[str], typer.Option("--output", "-o", help="Output merged PDF file path.")] = None,
    compress: Annotated[bool, typer.Option("--compress", "-c", help="Compress the output PDF.")] = False,
) -> None:
    """📄 Merge two PDF files into one."""
    from machineconfig.scripts.python.helpers.helpers_utils.pdf import merge_pdfs as impl
    impl(pdfs=pdfs, output=output, compress=compress)


def compress_pdf(
    pdf_input: Annotated[str, typer.Argument(..., help="Path to the input PDF file to compress.")],
    output: Annotated[Optional[str], typer.Option("--output", "-o", help="Output compressed PDF file path.")] = None,
    quality: Annotated[int, typer.Option("--quality", "-q", help="JPEG quality for image compression (0-100, 0=no change, 100=best).")] = 85,
    image_dpi: Annotated[int, typer.Option("--image-dpi", "-d", help="Target DPI for image resampling.")] = 0,
    compress_streams: Annotated[bool, typer.Option("--compress-streams", "-c", help="Compress uncompressed streams.")] = True,
    use_objstms: Annotated[bool, typer.Option("--object-streams", "-s", help="Use object streams for additional compression.")] = True,
) -> None:
    """📦 Compress a PDF file."""
    from machineconfig.scripts.python.helpers.helpers_utils.pdf import compress_pdf as impl
    impl(pdf_input=pdf_input, output=output, quality=quality, image_dpi=image_dpi, compress_streams=compress_streams, use_objstms=use_objstms)


def get_app() -> typer.Typer:
    app = typer.Typer(help="🛠️ utilities operations", no_args_is_help=True, add_help_option=True, add_completion=False)
    app.command(name="kill-process", no_args_is_help=False, help="⚔️ [k] Choose a process to kill")(kill_process)
    app.command(name="k", no_args_is_help=False, hidden=True)(kill_process)

    app.command("environment", no_args_is_help=False, help="📚 [v] NAVIGATE ENV/PATH variable with TUI")(tui_env)
    app.command("v", no_args_is_help=False, hidden=True)(tui_env)

    app.command(name="upgrade-packages", no_args_is_help=False, help="⬆️ [up] Upgrade project dependencies.")(upgrade_packages)
    app.command(name="up", no_args_is_help=False, hidden=True)(upgrade_packages)

    app.command(name="download", no_args_is_help=True, help="⬇️ [d] Download a file from a URL and optionally decompress it.")(download)
    app.command(name="d", no_args_is_help=True, hidden=True)(download)
    app.command(name="get-machine-specs", no_args_is_help=False, help="💻 [g] Get machine specifications.")(get_machine_specs)
    app.command(name="g", no_args_is_help=False, hidden=True)(get_machine_specs)
    app.command(name="init-project", no_args_is_help=False, help="🚀 [i] Initialize a project with a uv virtual environment and install dev packages.")(init_project)
    app.command(name="i", no_args_is_help=False, hidden=True)(init_project)
    app.command(name="edit", no_args_is_help=False, help="✏️ [e] Open a file in the default editor.")(edit_file_with_hx)
    app.command(name="e", no_args_is_help=False, hidden=True)(edit_file_with_hx)

    app.command(name="pdf-merge", no_args_is_help=True, help="📄 [pm] Merge two PDF files into one.")(merge_pdfs)
    app.command(name="pm", no_args_is_help=True, hidden=True)(merge_pdfs)
    app.command(name="pdf-compress", no_args_is_help=True, help="📦 [pc] Compress a PDF file.")(compress_pdf)
    app.command(name="pc", no_args_is_help=True, hidden=True)(compress_pdf)

    return app


def main():
    app = get_app()
    app()



"""shell"""

from pathlib import Path



def get_shell_profile_path() -> Path:
    import platform
    import subprocess
    from rich.console import Console
    from rich.panel import Panel
    system = platform.system()
    console = Console()
    if system == "Windows":
        result = subprocess.run(["pwsh", "-Command", "$PROFILE"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=False)
        if result.returncode == 0 and result.stdout.strip():
            profile_path = Path(result.stdout.strip())
        else:
            print(f"Command failed with return code {result.returncode}")
            print(f"stdout: {result.stdout}")
            print(f"stderr: {result.stderr}")
            raise ValueError(f"""Could not get profile path for Windows. Got stdout: {result.stdout}, stderr: {result.stderr}""")
    elif system == "Linux":
        profile_path = Path.home().joinpath(".bashrc")
    elif system == "Darwin":
        profile_path = Path.home().joinpath(".zshrc")
    else:
        raise ValueError(f"""Not implemented for this system {system}""")
    console.print(Panel(f"""🐚 SHELL PROFILE | Working with path: `{profile_path}`""", title="[bold blue]Shell Profile[/bold blue]", border_style="blue"))
    return profile_path


def get_nu_shell_profile_path() -> Path:
    import platform
    from rich.console import Console
    from rich.panel import Panel
    system = platform.system()
    console = Console()
    if system == "Windows":
        profile_path = Path.home().joinpath(r"AppData\Roaming\nushell")
    elif system == "Linux":
        profile_path = Path.home().joinpath(".config/nushell")
    elif system == "Darwin":
        profile_path = Path.home().joinpath("Library/Application Support/nushell")
    else:
        raise ValueError(f"""Not implemented for this system {system}""")
    console.print(Panel(f"""🐚 NU SHELL PROFILE | Working with path: `{profile_path}`""", title="[bold cyan]Nu Shell Profile[/bold cyan]", border_style="cyan"))
    return profile_path


def create_default_shell_profile() -> None:
    shell_profile_path = get_shell_profile_path()
    import platform
    import subprocess
    from rich.console import Console
    from rich.panel import Panel
    from machineconfig.utils.source_of_truth import CONFIG_ROOT
    from machineconfig.utils.path_extended import PathExtended
    system = platform.system()
    console = Console()
    if not shell_profile_path.exists():
        console.print(Panel(f"""🆕 PROFILE | Profile does not exist at `{shell_profile_path}`. Creating a new one.""", title="[bold blue]Profile[/bold blue]", border_style="blue"))
        shell_profile_path.parent.mkdir(parents=True, exist_ok=True)
        shell_profile_path.write_text("", encoding="utf-8")
    shell_profile = shell_profile_path.read_text(encoding="utf-8")
    from machineconfig.profile.create_helper import copy_assets_to_machine
    copy_assets_to_machine("settings")  # init.ps1 or init.sh live here
    copy_assets_to_machine("scripts")  # init scripts are going to reference those scripts.
    shell_name = ""
    if system == "Windows":
        shell_name = "pwsh"
        init_script = PathExtended(CONFIG_ROOT).joinpath("settings/shells/pwsh/init.ps1")
        source_line = f""". {str(init_script.collapseuser(placeholder="$HOME"))}"""
    elif system == "Linux":
        shell_name = "bash"
        init_script = PathExtended(CONFIG_ROOT).joinpath("settings/shells/bash/init.sh")
        source_line = f"""source {str(init_script.collapseuser(placeholder="$HOME"))}"""
    elif system == "Darwin":
        shell_name = "zsh"
        init_script = PathExtended(CONFIG_ROOT).joinpath("settings/shells/zsh/init.sh")
        source_line = f"""source {str(init_script.collapseuser(placeholder="$HOME"))}"""
    else:
        raise ValueError(f"""Not implemented for this system {system}""")

    was_shell_updated = False
    if source_line in shell_profile:
        console.print(Panel("🔄 PROFILE | Skipping init script sourcing - already present in profile", title="[bold blue]Profile[/bold blue]", border_style="blue"))
    else:
        console.print(Panel("📝 PROFILE | Adding init script sourcing to profile", title="[bold blue]Profile[/bold blue]", border_style="blue"))
        shell_profile += "\n" + source_line + "\n"
        if shell_name == "bash":
            result = subprocess.run(["cat", "/proc/version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=False)
            if result.returncode == 0 and result.stdout:
                version_info = result.stdout.lower()
                is_wsl = "microsoft" in version_info or "wsl" in version_info
                if is_wsl:
                    shell_profile += "\ncd $HOME"
                    console.print("📌 WSL detected - adding 'cd $HOME' to profile to avoid Windows filesystem")
                    # Sync shell history between Windows and WSL
                    # https://www.hanselman.com/blog/sharing-powershell-history-between-windows-and-wsl
                    shell_profile += """
# Sync shell history between Windows and WSL
export PROMPT_COMMAND="${PROMPT_COMMAND:+$PROMPT_COMMAND$'\\n'}history -a; history -c; history -r"
"""
        was_shell_updated = True
    if was_shell_updated:
        shell_profile_path.parent.mkdir(parents=True, exist_ok=True)
        shell_profile_path.write_text(shell_profile, encoding="utf-8")
        console.print(Panel("✅ Profile updated successfully", title="[bold blue]Profile[/bold blue]", border_style="blue"))


def create_nu_shell_profile() -> None:
    from rich.console import Console
    from rich.panel import Panel
    from machineconfig.utils.source_of_truth import CONFIG_ROOT
    from machineconfig.utils.path_extended import PathExtended
    console = Console()
    nu_profile_path = get_nu_shell_profile_path()
    config_dir = nu_profile_path
    config_file = config_dir.joinpath("config.nu")
    if not config_dir.exists():
        console.print(Panel(f"""🆕 NU SHELL CONFIG | Config directory does not exist at `{config_dir}`. Creating a new one.""", title="[bold cyan]Nu Shell Config[/bold cyan]", border_style="cyan"))
        config_dir.mkdir(parents=True, exist_ok=True)
    if not config_file.exists():
        console.print(Panel(f"""🆕 NU SHELL CONFIG | config.nu file does not exist at `{config_file}`. Creating a new one.""", title="[bold cyan]Nu Shell Config[/bold cyan]", border_style="cyan"))
        config_file.write_text("", encoding="utf-8")
    config_content = config_file.read_text(encoding="utf-8")
    from machineconfig.profile.create_helper import copy_assets_to_machine
    copy_assets_to_machine("settings")
    copy_assets_to_machine("scripts")
    init_script = PathExtended(CONFIG_ROOT).joinpath("settings/shells/nushell/init.nu")
    source_line = f"""use {str(init_script)}"""
    was_config_updated = False
    if source_line in config_content:
        console.print(Panel("🔄 NU SHELL CONFIG | Skipping init script sourcing - already present in config.nu", title="[bold cyan]Nu Shell Config[/bold cyan]", border_style="cyan"))
    else:
        console.print(Panel("📝 NU SHELL CONFIG | Adding init script sourcing to config.nu", title="[bold cyan]Nu Shell Config[/bold cyan]", border_style="cyan"))
        config_content += "\n" + source_line + "\n"
        was_config_updated = True
    if was_config_updated:
        config_dir.mkdir(parents=True, exist_ok=True)
        config_file.write_text(config_content, encoding="utf-8")
        console.print(Panel("✅ Nu shell config updated successfully", title="[bold cyan]Nu Shell Config[/bold cyan]", border_style="cyan"))


if __name__ == "__main__":
    pass



import typer
from typing import Annotated


def choose_zellij_session(
        name: Annotated[str | None, typer.Argument(help="Name of the Zellij session to attach to. If not provided, a list will be shown to choose from.")] = None,
        new_session: Annotated[bool, typer.Option("--new-session", "-n", help="Create a new Zellij session instead of attaching to an existing one.", show_default=True)] = False,
        kill_all: Annotated[bool, typer.Option("--kill-all", "-k", help="Kill all existing Zellij sessions before creating a new one.", show_default=True)] = False) -> None:
    """Choose a Zellij session to attach to."""
    from machineconfig.scripts.python.helpers.helpers_terminal.terminal_impl import choose_zellij_session as impl
    action, payload = impl(name=name, new_session=new_session, kill_all=kill_all)
    if action == "error":
        print(payload)
        raise typer.Exit()
    if action == "run_script" and payload:
        from machineconfig.utils.code import exit_then_run_shell_script
        exit_then_run_shell_script(payload, strict=True)


def get_session_tabs() -> list[tuple[str, str]]:
    """Get all Zellij session tabs."""
    from machineconfig.scripts.python.helpers.helpers_terminal.terminal_impl import get_session_tabs as impl
    result = impl()
    print(result)
    return result


def start_wt(layout_name: Annotated[str, typer.Argument(help="Layout name to start.")]) -> None:
    """Start a Windows Terminal layout by name."""
    from machineconfig.scripts.python.helpers.helpers_terminal.terminal_impl import start_wt as impl
    status, message = impl(layout_name=layout_name)
    if status == "error":
        typer.echo(message)
        raise typer.Exit(code=1)
    # cmd = f'powershell -ExecutionPolicy Bypass -File "./{layout_name}_layout.ps1"'
    # from machineconfig.utils.code import exit_then_run_shell_script
    # exit_then_run_shell_script(cmd, strict=True)


def get_app():
    app = typer.Typer(help="🖥️ Terminal utilities", no_args_is_help=True, add_help_option=True, add_completion=False)
    app.command(name="attach-to-zellij", no_args_is_help=False, help=choose_zellij_session.__doc__, short_help="[z] Choose a Zellij session to attach to")(choose_zellij_session)
    app.command(name="z", hidden=True, no_args_is_help=False, help=choose_zellij_session.__doc__)(choose_zellij_session)

    app.command(name="start-wt", no_args_is_help=True, help=start_wt.__doc__, short_help="[w] Start a Windows Terminal layout by name.")(start_wt)
    app.command(name="w", hidden=True, no_args_is_help=True, help=start_wt.__doc__)(start_wt)

    app.command(name="get-session-tabs", no_args_is_help=False, help=get_session_tabs.__doc__, short_help="[zt] Get all Zellij session tabs.")(get_session_tabs)
    app.command(name="zt", hidden=True, no_args_is_help=False, help=get_session_tabs.__doc__)(get_session_tabs)
    return app

def main():
    app = get_app()
    app()


if __name__ == "__main__":
    main()





import platform
from typing import TYPE_CHECKING, Optional

from machineconfig.utils.installer_utils.installer_helper import install_deb_package, download_and_prepare
from machineconfig.utils.installer_utils.installer_locator_utils import find_move_delete_linux, find_move_delete_windows
from machineconfig.utils.installer_utils.github_release_bulk import (
    get_repo_name_from_url,
    fetch_github_release_data,
    extract_release_info,
    AssetInfo,
)
from machineconfig.utils.path_extended import PathExtended
from machineconfig.utils.source_of_truth import INSTALL_VERSION_ROOT

if TYPE_CHECKING:
    from rich.console import Console

SUPPORTED_GITHUB_HOSTS = {"github.com", "www.github.com"}


def _format_size(size_bytes: int) -> str:
    if size_bytes <= 0:
        return "0 B"
    units = ("B", "KiB", "MiB", "GiB", "TiB")
    value = float(size_bytes)
    index = 0
    while value >= 1024 and index < len(units) - 1:
        value /= 1024
        index += 1
    return f"{value:.1f} {units[index]}"


def _derive_tool_name(repo_name: str, asset_name: Optional[str]) -> Optional[str]:
    repo_segment = repo_name.split("/", maxsplit=1)[-1]
    repo_clean = repo_segment.replace(".git", "").lower()
    repo_filtered = "".join(char for char in repo_clean if char.isalnum())
    if repo_filtered:
        return repo_filtered
    if asset_name is None:
        return None
    asset_clean = asset_name.lower()
    asset_filtered = "".join(char for char in asset_clean if char.isalnum())
    if asset_filtered:
        return asset_filtered
    return None


def _finalize_install(repo_name: str, asset_name: Optional[str], version: str, extracted_path: PathExtended, console: "Console") -> None:
    from rich.panel import Panel
    if extracted_path.suffix == ".deb":
        install_deb_package(extracted_path)
        tool_name_deb = _derive_tool_name(repo_name, asset_name)
        if tool_name_deb is not None:
            INSTALL_VERSION_ROOT.joinpath(tool_name_deb).parent.mkdir(parents=True, exist_ok=True)
            INSTALL_VERSION_ROOT.joinpath(tool_name_deb).write_text(version, encoding="utf-8")
        console.print(Panel(f"Installed Debian package for [green]{tool_name_deb}[/green]", title="✅ Complete", border_style="green"))
        return
    system_name = platform.system()
    tool_name = _derive_tool_name(repo_name, asset_name)
    rename_target = f"{tool_name}.exe" if system_name == "Windows" else tool_name
    try:
        if system_name == "Windows":
            installed_path = find_move_delete_windows(downloaded_file_path=extracted_path, tool_name=tool_name, delete=True, rename_to=rename_target)
        elif system_name in {"Linux", "Darwin"}:
            installed_path = find_move_delete_linux(downloaded=extracted_path, tool_name=tool_name, delete=True, rename_to=rename_target)
        else:
            console.print(Panel(f"Unsupported operating system: {system_name}", title="❌ Error", border_style="red"))
            return None
    except IndexError:
        if system_name == "Windows":
            installed_path = find_move_delete_windows(downloaded_file_path=extracted_path, tool_name=None, delete=True, rename_to=rename_target)
        elif system_name in {"Linux", "Darwin"}:
            installed_path = find_move_delete_linux(downloaded=extracted_path, tool_name="", delete=True, rename_to=rename_target)
        else:
            raise
    if tool_name is not None:
        INSTALL_VERSION_ROOT.joinpath(tool_name).parent.mkdir(parents=True, exist_ok=True)
        INSTALL_VERSION_ROOT.joinpath(tool_name).write_text(version, encoding="utf-8")
    console.print(Panel(f"Installed [green]{tool_name}[/green] to {installed_path}\nVersion: {version}", title="✅ Complete", border_style="green"))


def install_from_github_url(github_url: str) -> None:
    from machineconfig.utils.options import choose_from_options
    from rich.console import Console
    from rich.panel import Panel

    console = Console()
    repo_info = get_repo_name_from_url(github_url)
    if repo_info is None:
        console.print(Panel(f"Invalid GitHub URL: {github_url}", title="❌ Error", border_style="red"))
        return None
    owner, repo = repo_info
    repo_name = f"{owner}/{repo}"
    console.print(Panel(f"Fetching latest release for [green]{repo_name}[/green]", title="🌐 GitHub", border_style="blue"))
    release_raw = fetch_github_release_data(owner, repo)
    if not release_raw:
        console.print(Panel("No releases available for this repository.", title="❌ Error", border_style="red"))
        return None
    
    release_info = extract_release_info(release_raw)
    if not release_info:
        console.print(Panel("Failed to parse release information.", title="❌ Error", border_style="red"))
        return None
    
    assets = release_info["assets"]
    if not assets:
        console.print(Panel("No downloadable assets found in the latest release.", title="❌ Error", border_style="red"))
        return None
    binary_assets = assets
    selection_pool = binary_assets if binary_assets else assets
    if not selection_pool:
        console.print(Panel("No assets available for installation.", title="❌ Error", border_style="red"))
        return None
    
    # First pass: collect all formatted data and calculate column widths
    asset_data = []
    for asset in selection_pool:
        name = asset["name"]
        download_url = asset["browser_download_url"]
        if name == "" or download_url == "":
            continue
        size = asset["size"]
        download_count = asset.get("download_count", 0)
        created_at = asset.get("created_at", "")
        
        # Format each field
        size_str = f"[{_format_size(size)}]"
        downloads_str = f"{download_count:,}"
        date_str = created_at.split("T")[0] if created_at else "N/A"
        
        asset_data.append({
            "name": name,
            "size_str": size_str,
            "downloads_str": downloads_str,
            "date_str": date_str,
            "asset": asset
        })
    
    # Calculate maximum widths for alignment
    max_name_len = max(len(item["name"]) for item in asset_data) if asset_data else 0
    max_size_len = max(len(item["size_str"]) for item in asset_data) if asset_data else 0
    max_downloads_len = max(len(item["downloads_str"]) for item in asset_data) if asset_data else 0
    
    # Second pass: build aligned labels
    options_map: dict[str, AssetInfo] = {}
    for item in asset_data:
        name_padded = item["name"].ljust(max_name_len)
        size_padded = item["size_str"].ljust(max_size_len)
        downloads_padded = item["downloads_str"].rjust(max_downloads_len)
        
        label = f"{name_padded} {size_padded} | #🔽 {downloads_padded} | 📅 {item['date_str']}"
        options_map[label] = item["asset"]
    
    if not options_map:
        console.print(Panel("Release assets lack download URLs.", title="❌ Error", border_style="red"))
        return None
    selection_label = choose_from_options(options=list(options_map.keys()), msg="Select a release asset", multi=False, header="📦 GitHub Release Assets", tv=True)
    selected_asset = options_map[selection_label]
    download_url_value = selected_asset["browser_download_url"]
    asset_name_value = selected_asset["name"]
    if download_url_value == "":
        console.print(Panel("Selected asset lacks a download URL.", title="❌ Error", border_style="red"))
        return None
    asset_name = asset_name_value if asset_name_value != "" else "github_binary"
    version = release_info["tag_name"] if release_info["tag_name"] != "" else "latest"
    console.print(Panel(f"Downloading [cyan]{asset_name}[/cyan]", title="⬇️ Download", border_style="magenta"))
    extracted_path = download_and_prepare(download_url_value)
    _finalize_install(repo_name=repo_name, asset_name=asset_name, version=version, extracted_path=extracted_path, console=console)


def install_from_binary_url(binary_url: str) -> None:
    from rich.console import Console
    # from rich.panel import Panel
    console = Console()
    # parsed = urlparse(binary_url)
    # asset_candidate = parsed.path.split("/")[-1] if parsed.path else ""
    # asset_name = asset_candidate if asset_candidate != "" else "binary_asset"
    # host = parsed.netloc if parsed.netloc != "" else "remote host"
    # console.print(Panel(f"Downloading from [green]{binary_url}[/green]", title="⬇️ Download", border_style="magenta"))
    extracted_path = download_and_prepare(binary_url)
    _finalize_install(repo_name="", asset_name=None, version="latest", extracted_path=extracted_path, console=console)



"""Pure Python implementation for ftpx command - no typer dependencies."""

from pathlib import Path
from typing import Optional


def ftpx(source: str, target: str, recursive: bool, zipFirst: bool, cloud: bool, overwrite_existing: bool) -> None:
    """File transfer utility through SSH."""
    if target == "wsl" or source == "wsl":
        _handle_wsl_transfer(source=source, target=target, overwrite_existing=overwrite_existing, )
        return
    elif source == "win" or target == "win":
        _handle_win_transfer(source=source, target=target, overwrite_existing=overwrite_existing, windows_username=None)
        return

    from rich.console import Console
    from rich.panel import Panel

    console = Console()

    console.print(
        Panel(
            "\n".join([
                "🚀 FTP File Transfer",
                "📋 Starting transfer process...",
            ]),
            title="Transfer Initialisation",
            border_style="blue",
            padding=(1, 2),
        )
    )

    resolved_source, resolved_target, machine, source_is_remote = _resolve_paths(source=source, target=target)

    from machineconfig.utils.accessories import pprint
    pprint({"source": str(resolved_source), "target": str(resolved_target), "machine": machine}, "CLI Resolution")

    ssh = _create_ssh_connection(machine=machine, console=console)

    if cloud:
        received_file = _handle_cloud_transfer(ssh=ssh, resolved_source=resolved_source, resolved_target=resolved_target, console=console)
    else:
        received_file = _handle_direct_transfer(
            ssh=ssh, resolved_source=resolved_source, resolved_target=resolved_target,
            source_is_remote=source_is_remote, zipFirst=zipFirst, recursive=recursive,
            overwrite_existing=overwrite_existing, console=console
        )

    if source_is_remote and received_file is not None:
        from machineconfig.utils.path_extended import PathExtended
        if isinstance(received_file, PathExtended):
            console.print(
                Panel(
                    "\n".join([
                        "📁 File Received",
                        f"Parent: [cyan]{repr(received_file.parent)}[/cyan]",
                        f"File: [cyan]{repr(received_file)}[/cyan]",
                    ]),
                    title="Transfer Result",
                    border_style="green",
                    padding=(1, 2),
                )
            )
    console.print(
        Panel(
            "File transfer process finished successfully",
            title="✅ Transfer Complete",
            border_style="green",
            padding=(1, 2),
        )
    )


def _handle_wsl_transfer(source: str, target: str, overwrite_existing: bool) -> None:
    """Handle WSL transfer when inside Windows."""
    from machineconfig.utils.ssh_utils.wsl import copy_when_inside_windows
    if target == "wsl":
        target_obj = Path(source).expanduser().absolute().relative_to(Path.home())
        source_obj = target_obj
    else:
        source_obj = Path(target).expanduser().absolute().relative_to(Path.home())
        target_obj = source_obj
    copy_when_inside_windows(source_obj, target_obj, overwrite_existing)


def _handle_win_transfer(source: str, target: str, overwrite_existing: bool, windows_username: str | None) -> None:
    """Handle Windows transfer when inside WSL."""
    from machineconfig.utils.ssh_utils.wsl import copy_when_inside_wsl
    if source == "win":
        source_obj = Path(target).expanduser().absolute().relative_to(Path.home())
        target_obj = source_obj
    else:
        target_obj = Path(source).expanduser().absolute().relative_to(Path.home())
        source_obj = target_obj
    copy_when_inside_wsl(source_obj, target_obj, overwrite_existing, windows_username=windows_username)


def _resolve_paths(source: str, target: str) -> tuple[Optional[str], Optional[str], str, bool]:
    """Resolve source and target paths, determine machine and direction."""
    from machineconfig.utils.path_extended import PathExtended
    from machineconfig.scripts.python.helpers.helpers_cloud.helpers2 import ES

    resolved_source: Optional[str] = None
    resolved_target: Optional[str] = None
    machine: str = ""
    source_is_remote: bool = False

    if ":" in source and (source[1] != ":" if len(source) > 1 else True):
        source_is_remote = True
        source_parts = source.split(":")
        machine = source_parts[0]
        if len(source_parts) > 1 and source_parts[1] == ES:
            if target == ES:
                raise ValueError(f"""
┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
┃ ❌ Configuration Error
┃    Cannot use expand symbol `{ES}` in both source and target
┃    This creates a cyclical inference dependency
┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━""")
            else:
                target_path_obj = PathExtended(target).expanduser().absolute()
            resolved_source = target_path_obj.collapseuser().as_posix()
            resolved_target = target
        else:
            resolved_source = ":".join(source.split(":")[1:])
            if target == ES:
                resolved_target = None
            else:
                resolved_target = PathExtended(target).expanduser().absolute().as_posix()

    elif ":" in target and (target[1] != ":" if len(target) > 1 else True):
        source_is_remote = False
        target_parts = target.split(":")
        machine = target_parts[0]
        if len(target_parts) > 1 and target_parts[1] == ES:
            if source == ES:
                raise ValueError(f"""
┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
┃ ❌ Configuration Error
┃    Cannot use expand symbol `{ES}` in both source and target
┃    This creates a cyclical inference dependency
┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━""")
            else:
                resolved_source = source
            resolved_target = None
        else:
            resolved_target = ":".join(target.split(":")[1:])
            if source == ES:
                resolved_source = None
            else:
                resolved_source = PathExtended(source).expanduser().absolute().as_posix()
    else:
        raise ValueError("""
┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
┃ ❌ Path Error
┃    Either source or target must be a remote path
┃    Format should be: machine:path
┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━""")

    return resolved_source, resolved_target, machine, source_is_remote


def _create_ssh_connection(machine: str, console: "Console") -> "SSH":  # type: ignore[name-defined]
    """Create SSH connection, handling authentication."""
    from machineconfig.utils.ssh import SSH
    from paramiko.ssh_exception import AuthenticationException  # type: ignore
    from rich.panel import Panel

    try:
        ssh = SSH(host=rf"{machine}", username=None, hostname=None, ssh_key_path=None, password=None, port=22, enable_compression=True)
    except AuthenticationException:
        console.print(
            Panel(
                "\n".join([
                    "🔑 Authentication failed. Trying manual authentication...",
                    "⚠️  Ensure that the username is provided correctly; only password prompts are handled here.",
                ]),
                title="Authentication Required",
                border_style="yellow",
                padding=(1, 2),
            )
        )
        import getpass
        pwd = getpass.getpass()
        ssh = SSH(host=rf"{machine}", username=None, hostname=None, ssh_key_path=None, password=pwd, port=22, enable_compression=True)

    return ssh


def _handle_cloud_transfer(ssh: "SSH", resolved_source: Optional[str], resolved_target: Optional[str], console: "Console") -> Optional["PathExtended"]:  # type: ignore[name-defined]
    """Handle cloud transfer mode."""
    from machineconfig.utils.path_extended import PathExtended
    from rich.panel import Panel

    console.print(
        Panel.fit(
            "☁️  Cloud transfer mode — uploading from remote to cloud...",
            title="Cloud Upload",
            border_style="cyan",
        )
    )
    ssh.run_shell_cmd_on_remote(command=f"cloud_copy {resolved_source} :^", verbose_output=True, description="Uploading from remote to the cloud.", strict_stderr=False, strict_return_code=False)
    console.print(
        Panel.fit(
            "⬇️  Cloud transfer mode — downloading from cloud to local...",
            title="Cloud Download",
            border_style="cyan",
        )
    )
    ssh.run_shell_cmd_on_local(command=f"cloud_copy :^ {resolved_target}")
    return PathExtended(resolved_target)  # type: ignore


def _handle_direct_transfer(ssh: "SSH", resolved_source: Optional[str], resolved_target: Optional[str], source_is_remote: bool, zipFirst: bool, recursive: bool, overwrite_existing: bool, console: "Console") -> Optional["PathExtended"]:  # type: ignore[name-defined]
    """Handle direct SSH transfer."""
    from rich.panel import Panel

    if source_is_remote:
        if resolved_source is None:
            print("❌ Path Error: Source must be a remote path (machine:path)")
            return None
        target_display = resolved_target or "<auto>"
        console.print(
            Panel(
                "\n".join([
                    "📥 Transfer Mode: Remote → Local",
                    f"Source: [cyan]{resolved_source}[/cyan]",
                    f"Target: [cyan]{target_display}[/cyan]",
                    f"Options: {'ZIP compression' if zipFirst else 'No compression'}, {'Recursive' if recursive else 'Non-recursive'}",
                ]),
                title="Transfer Details",
                border_style="cyan",
                padding=(1, 2),
            )
        )
        received_file = ssh.copy_to_here(source=resolved_source, target=resolved_target, compress_with_zip=zipFirst, recursive=recursive)
    else:
        assert resolved_source is not None, "❌ Path Error: Target must be a remote path (machine:path)"
        target_display = resolved_target or "<auto>"
        console.print(
            Panel(
                "\n".join([
                    "📤 Transfer Mode: Local → Remote",
                    f"Source: [cyan]{resolved_source}[/cyan]",
                    f"Target: [cyan]{target_display}[/cyan]",
                    f"Options: {'ZIP compression' if zipFirst else 'No compression'}, {'Recursive' if recursive else 'Non-recursive'}",
                ]),
                title="Transfer Details",
                border_style="cyan",
                padding=(1, 2),
            )
        )
        received_file = ssh.copy_from_here(source_path=resolved_source, target_rel2home=resolved_target, compress_with_zip=zipFirst, recursive=recursive, overwrite_existing=overwrite_existing)

    return received_file




from typing import Literal


def list_available_scripts(where: Literal["all", "a", "private", "p", "public", "b", "library", "l", "dynamic", "d", "custom", "c"]) -> None:
    from pathlib import Path
    from machineconfig.utils.source_of_truth import CONFIG_ROOT, LIBRARY_ROOT, DEFAULTS_PATH

    private_root = Path.home().joinpath("dotfiles/scripts")
    public_root = CONFIG_ROOT.joinpath("scripts")
    library_root = LIBRARY_ROOT.joinpath("jobs", "scripts")

    def get_custom_roots() -> list[Path]:
        custom_roots: list[Path] = []
        if DEFAULTS_PATH.is_file():
            from configparser import ConfigParser
            config = ConfigParser()
            config.read(DEFAULTS_PATH)
            if config.has_section("general") and config.has_option("general", "scripts"):
                custom_dirs = config.get("general", "scripts").split(",")
                for custom_dir in custom_dirs:
                    custom_path = Path(custom_dir.strip()).expanduser().resolve()
                    if custom_path.is_dir():
                        custom_roots.append(custom_path)
        return custom_roots

    locations: dict[str, Path | str] = {}
    match where:
        case "all" | "a":
            locations = {"private": private_root, "public": public_root, "library": library_root}
            for idx, custom in enumerate(get_custom_roots()): locations[f"custom_{idx}"] = custom
            locations["dynamic"] = "https://github.com/thisismygitrepo/machineconfig/tree/main/src/machineconfig/jobs/scripts_dynamic"
        case "private" | "p":
            locations = {"private": private_root}
        case "public" | "b":
            locations = {"public": public_root}
        case "library" | "l":
            locations = {"library": library_root}
        case "dynamic" | "d":
            locations = {"dynamic": "https://github.com/thisismygitrepo/machineconfig/tree/main/src/machineconfig/jobs/scripts_dynamic"}
        case "custom" | "c":
            for idx, custom in enumerate(get_custom_roots()): locations[f"custom_{idx}"] = custom

    def _print_files_by_type(files: list[str]) -> None:
        categories: dict[str, list[str]] = {".py": [], ".sh": [], ".ps1": [], ".cmd": [], ".bat": [], "other": []}
        for f in files:
            ext = Path(f).suffix.lower() if "." in str(f) else ""
            if ext in categories: categories[ext].append(str(f))
            else: categories["other"].append(str(f))
        for ext, cat_files in categories.items():
            if cat_files:
                label = ext if ext else "other"
                print(f"    [{label}]")
                for cf in sorted(cat_files): print(f"      • {cf}")

    for loc_name, loc_path in locations.items():
        print(f"\n📁 {loc_name.upper()} ({loc_path}):")
        print("-" * 60)
        if isinstance(loc_path, Path):
            if loc_path.is_dir():
                files = [f for f in loc_path.rglob("*") if f.is_file() and f.suffix in (".py", ".sh", ".ps1", ".bat", ".cmd", "")]
                if files:
                    relative_files = [str(f.relative_to(loc_path)) for f in files]
                    _print_files_by_type(relative_files)
                else:
                    print("  (empty)")
            else:
                print("  ⚠️  Directory does not exist")
        else:
            api_url = "https://api.github.com/repos/thisismygitrepo/machineconfig/contents/src/machineconfig/jobs/scripts_dynamic"
            try:
                import requests
                response = requests.get(api_url, timeout=10)
                if response.status_code == 200:
                    items = response.json()
                    files = [item["name"] for item in items if item["type"] == "file"]
                    _print_files_by_type(files)
                else:
                    print(f"  ⚠️  Could not fetch from GitHub (status: {response.status_code})")
            except Exception as e:
                print(f"  ⚠️  Could not fetch from GitHub: {e}")




"""run python/shell scripts from pre-defined directorys, or, run command/file in the machineconfig environment

Recursively Searched Predefined Directories:


* 'private' : $HOME/dotfiles/scripts

* 'public'  : $HOME/.config/machineconfig/scripts

* 'library' : $MACHINECONFIG_LIBRARY_ROOT/jobs/scripts

* 'dynamic' : fetched from GitHub repository on the fly (relies on latest commit, rather than the version currently installed)

* 'custom'  : custom directories from comma separated entry 'scripts' under 'general' section @ ~/dotfiles/machineconfig/defaults.ini

"""


import typer
from typing import Annotated, Optional, Literal


def run_py_script(name: Annotated[str, typer.Argument(help="Name of script to run, e.g., 'a' for a.py, or command to execute")] = "",
                  where: Annotated[Literal["all", "a", "private", "p", "public", "b", "library", "l", "dynamic", "d", "custom", "c"], typer.Option("--where", "-w", help="Where to look for the script")] = "all",
                  interactive: Annotated[bool, typer.Option(..., "--interactive", "-i", help="Interactive selection of scripts to run")] = False,
                  command: Annotated[Optional[bool], typer.Option(..., "--command", "-c", help="Run as command")] = False,
                  list_scripts: Annotated[bool, typer.Option(..., "--list", "-l", help="List available scripts in all locations")] = False,
                ) -> None:
    from pathlib import Path

    if list_scripts:
        from machineconfig.scripts.python.helpers.helpers_search.script_help import list_available_scripts
        list_available_scripts(where=where)
        return

    if not interactive and not name:
        typer.echo("❌ ERROR: You must provide a script name or use --interactive option to select a script.")
        raise typer.Exit(code=1)
    target_file: Optional[Path] = None

    if where in ["dynamic", "d"]:
        # src/machineconfig/jobs/scripts/python_scripts/a.py
        if "." in name: resolved_names: list[str] = [name]
        else: resolved_names = [f"{name}{a_suffix}" for a_suffix in [".py", ".sh", "", ".ps1", ".bat", ".cmd"]]
        urls = [f"""https://raw.githubusercontent.com/thisismygitrepo/machineconfig/refs/heads/main/src/machineconfig/jobs/scripts_dynamic/{a_resolved_name}""" for a_resolved_name in resolved_names]
        for a_url in urls:
            try:
                print(f"Fetching temporary script from {a_url} ...")
                import requests
                response = requests.get(a_url)
                if response.status_code != 200:
                    print(f"❌ ERROR: Could not fetch script '{name}.py' from repository. Status Code: {response.status_code}")
                    raise RuntimeError(f"Could not fetch script '{name}.py' from repository.")
                script_content = response.text
                target_file = Path.home().joinpath("tmp_results", "tmp_scripts", "python", f"{name}.py")
                target_file.parent.mkdir(parents=True, exist_ok=True)
                target_file.write_text(script_content, encoding="utf-8")
            except Exception as _e:
                pass

    if command:
        exec(name)
        return
    if target_file is None and Path(name).is_file():
        if name.endswith(".py"):
            import machineconfig
            import subprocess
            import sys
            subprocess.run([sys.executable, name], cwd=machineconfig.__path__[0])
            return
        else:
            raise RuntimeError(f"File '{name}' is not a python (.py) file.")

    from machineconfig.utils.source_of_truth import CONFIG_ROOT, LIBRARY_ROOT, DEFAULTS_PATH
    private_root = Path.home().joinpath("dotfiles/scripts")  # local directory
    public_root = CONFIG_ROOT.joinpath("scripts")  # local machineconfig directory
    library_root = LIBRARY_ROOT.joinpath("jobs", "scripts")

    def get_custom_roots() -> list[Path]:
        custom_roots: list[Path] = []
        if DEFAULTS_PATH.is_file():
            from configparser import ConfigParser
            config = ConfigParser()
            config.read(DEFAULTS_PATH)
            if config.has_section("general") and config.has_option("general", "scripts"):
                custom_dirs = config.get("general", "scripts").split(",")
                for custom_dir in custom_dirs:
                    custom_path = Path(custom_dir.strip()).expanduser().resolve()
                    if custom_path.is_dir():
                        custom_roots.append(custom_path)
        return custom_roots

    roots: list[Path] = []
    match where:
        case "all" | "a":
            roots = [private_root, public_root, library_root] + get_custom_roots()
        case "private" | "p":
            roots = [private_root]
        case "public" | "b":
            roots = [public_root]
        case "library" | "l":
            roots = [library_root]
        case "dynamic" | "d":
            roots = []
        case "custom" | "c":
            roots = get_custom_roots()

    suffixes: list[str]
    if "." in name:
        suffixes = [""]
    else:
        import platform
        if platform.system() == "Windows":
            suffixes = [".py", ".bat", ".cmd", ".ps1"]
        elif platform.system() == "Darwin" or platform.system() == "Linux":
            suffixes = [""]  # files without suffix could be shell scripts, and that already cover files with .sh suffix without duplications
        else:
            suffixes = [".py"]

    # Finding target file
    potential_matches: list[Path] = []
    if target_file is None:
        for a_root in roots:
            for a_suffix in suffixes:
                if a_root.joinpath(f"{name}{a_suffix}").is_file():
                    target_file = a_root.joinpath(f"{name}{a_suffix}")
                    break  # perfect match
                potential_matches += [a_file for a_file in a_root.rglob(f"*{name}*{a_suffix}") if a_file.is_file()]

    if target_file is None:
        if len(potential_matches) == 1:
            target_file = potential_matches[0]
        elif len(potential_matches) == 0:
            print(f"❌ ERROR: Could not find script '{name}'.")
            print("Searched in:")
            for r in roots:
                print(f"  - {r}")
            raise typer.Exit(code=1)
        else:
            print(f"Warning: Could not find script '{name}'. Checked {len(potential_matches)} candidate files, trying interactively:")
            from machineconfig.utils.options import choose_from_options
            options = [str(p) for p in potential_matches]
            chosen_file_part = choose_from_options(options, multi=False, msg="Select the script to run:", tv=True, preview="bat")
            target_file = Path(chosen_file_part)

    print(f"✅ Found script at: {target_file}")
    if target_file.suffix == ".py":
        from machineconfig.utils.code import get_uv_command_executing_python_file, exit_then_run_shell_script
        shell_script = get_uv_command_executing_python_file(python_file=str(target_file), uv_project_dir=None, uv_with=None, prepend_print=False)
        exit_then_run_shell_script(script=shell_script)
    else:
        from machineconfig.utils.code import run_shell_file
        run_shell_file(script_path=str(target_file), clean_env=False)


def copy_script_to_local(name: Annotated[str, typer.Argument(help="Name of the temporary python script to copy, e.g., 'a' for a.py")],
                         alias: Annotated[Optional[str], typer.Option("--alias", "-a", help="Whether to create call it a different name locally")] = None
                         ) -> None:
    """
    Copy a temporary python script stored in machineconfig/scripts/python/helpers/tmp_py_scripts to the local machine.
    """
    url = f"""https://raw.githubusercontent.com/thisismygitrepo/machineconfig/refs/heads/main/src/machineconfig/scripts/python/helpers/tmp_py_scripts/{name}.py"""
    import requests
    response = requests.get(url)
    if response.status_code != 200:
        print(f"❌ ERROR: Could not fetch script '{name}.py' from repository. Status Code: {response.status_code}")
        raise RuntimeError(f"Could not fetch script '{name}.py' from repository.")
    script_content = response.text
    from machineconfig.utils.source_of_truth import CONFIG_ROOT
    local_path = CONFIG_ROOT.joinpath(f"scripts_python/{alias or name}.py")
    with open(local_path, "w") as f:
        f.write(script_content)
    print(f"✅ Script '{name}.py' has been copied to '{local_path}'.")


def get_app():
    app = typer.Typer(
        name="run-tmp-script",
        help="Helper to run temporary python scripts stored in machineconfig/scripts/python/helpers/tmp_py_scripts",
        no_args_is_help=True,
    )
    from machineconfig.jobs.scripts_dynamic import a
    app.command()(a.main)
    return app



"""Pure Python implementation for fire_jobs route command - no typer dependencies."""

from typing import Optional, Callable
from pathlib import Path
from machineconfig.scripts.python.helpers.helpers_fire_command.fire_jobs_args_helper import FireJobArgs

RandStrFunc = Callable[[int], str]


def route(args: FireJobArgs, fire_args: str) -> None:
    """Route execution based on args configuration."""
    from machineconfig.utils.path_helper import get_choice_file
    from machineconfig.utils.accessories import get_repo_root, randstr

    choice_file = get_choice_file(args.path, suffixes=None)
    repo_root = get_repo_root(choice_file)
    print(f"💾 Selected file: {choice_file}.\nRepo root: {repo_root}")

    if args.marimo:
        _handle_marimo(choice_file=choice_file, repo_root=repo_root, randstr_func=randstr)
        return

    kwargs_dict = _prepare_kwargs(args=args, choice_file=choice_file)
    choice_function = _choose_function(args=args, choice_file=choice_file, kwargs_dict=kwargs_dict)
    if isinstance(choice_function, tuple):
        choice_function, choice_file, kwargs_dict = choice_function

    command = _build_command(args=args, choice_file=choice_file, choice_function=choice_function, kwargs_dict=kwargs_dict, repo_root=repo_root, fire_args=fire_args, randstr_func=randstr)
    command = _apply_command_modifiers(args=args, command=command, choice_file=choice_file, repo_root=repo_root, randstr_func=randstr)

    from machineconfig.utils.code import exit_then_run_shell_script
    exit_then_run_shell_script(script=command, strict=False)


def _handle_marimo(choice_file: Path, repo_root: Optional[Path], randstr_func: RandStrFunc) -> None:
    """Handle marimo notebook launch."""
    print(f"🧽 Preparing to launch Marimo notebook for `{choice_file}`...")
    tmp_dir = Path.home().joinpath(f"tmp_results/tmp_scripts/marimo/{choice_file.stem}_{randstr_func(10)}")
    tmp_dir.mkdir(parents=True, exist_ok=True)
    script = f"""
cd {tmp_dir}
uv run --python 3.14 --with marimo marimo convert {choice_file} -o marimo_nb.py
uv run --project {repo_root} --with marimo marimo edit --host 0.0.0.0 marimo_nb.py
"""
    from machineconfig.utils.code import exit_then_run_shell_script
    print(f"🚀 Launching Marimo notebook for `{choice_file}`...")
    exit_then_run_shell_script(script)


def _prepare_kwargs(args: FireJobArgs, choice_file: Path) -> dict[str, object]:
    """Prepare kwargs dict from args."""
    if choice_file.suffix == ".py":
        from machineconfig.scripts.python.helpers.helpers_fire_command.fire_jobs_args_helper import extract_kwargs
        return extract_kwargs(args)
    return {}


def _choose_function(args: FireJobArgs, choice_file: Path, kwargs_dict: dict[str, object]) -> Optional[str] | tuple[Optional[str], Path, dict[str, object]]:
    """Choose function to run, possibly interactively."""
    if args.choose_function:
        from machineconfig.scripts.python.helpers.helpers_fire_command.fire_jobs_route_helper import choose_function_or_lines
        choice_function, choice_file, kwargs_dict = choose_function_or_lines(choice_file, kwargs_dict)
        return (choice_function, choice_file, kwargs_dict)
    return args.function


def _build_command(args: FireJobArgs, choice_file: Path, choice_function: Optional[str], kwargs_dict: dict[str, object], repo_root: Optional[Path], fire_args: str, randstr_func: RandStrFunc) -> str:
    """Build the execution command."""
    if choice_file.suffix == ".py":
        exe_line = _build_python_exe_line(args=args, choice_file=choice_file, repo_root=repo_root)
        choice_file_adjusted = _adjust_choice_file(args=args, choice_file=choice_file, repo_root=repo_root)

        if args.script or (args.debug and args.choose_function):
            choice_file = _create_import_script(choice_file=choice_file, choice_function=choice_function, kwargs_dict=kwargs_dict, repo_root=repo_root, randstr_func=randstr_func)
            choice_file_adjusted = str(choice_file)

        return _build_final_command(args=args, exe_line=exe_line, choice_file=choice_file, choice_file_adjusted=choice_file_adjusted, choice_function=choice_function, fire_args=fire_args)
    elif choice_file.suffix in (".ps1", ".sh"):
        return f". {choice_file}"
    elif choice_file.suffix == "":
        return str(choice_file)
    else:
        raise NotImplementedError(f"File type {choice_file.suffix} not supported, in the sense that I don't know how to fire it.")


def _build_python_exe_line(args: FireJobArgs, choice_file: Path, repo_root: Optional[Path]) -> str:
    """Build Python execution line."""
    module_line = "-m" if args.module else ""
    with_project = f"--project {repo_root} " if repo_root is not None else ""
    interactive_line = "-i" if args.interactive else ""

    if args.interactive:
        from machineconfig.utils.ve import get_ve_path_and_ipython_profile
        _ve_root_from_file, ipy_profile = get_ve_path_and_ipython_profile(init_path=choice_file)
        if ipy_profile is None:
            ipy_profile = "default"
        ipython_line = f"--no-banner --profile {ipy_profile} "
    else:
        ipython_line = ""

    if args.streamlit:
        from machineconfig.scripts.python.helpers.helpers_fire_command.fire_jobs_route_helper import get_command_streamlit
        interpreter_line = get_command_streamlit(choice_file=choice_file, environment=args.environment, repo_root=repo_root)
    elif args.jupyter:
        interpreter_line = "jupyter-lab"
    else:
        interpreter_line = "python" if not args.interactive else "ipython"

    return f"uv run {with_project} {interpreter_line} {interactive_line} {module_line} {ipython_line}"


def _adjust_choice_file(args: FireJobArgs, choice_file: Path, repo_root: Optional[Path]) -> str:
    """Adjust choice file path for module mode."""
    if args.module and choice_file.suffix == ".py":
        if repo_root is not None:
            return ".".join(Path(choice_file).relative_to(repo_root).parts).replace(".py", "")
        else:
            return ".".join(Path(choice_file).relative_to(Path.cwd()).parts).replace(".py", "")
    return str(choice_file)


def _create_import_script(choice_file: Path, choice_function: Optional[str], kwargs_dict: dict[str, object], repo_root: Optional[Path], randstr_func: RandStrFunc) -> Path:
    """Create a script that imports the module and calls the function."""
    from machineconfig.scripts.python.helpers.helpers_fire_command.file_wrangler import get_import_module_code, wrap_import_in_try_except
    from machineconfig.utils.meta import lambda_to_python_script
    from machineconfig.utils.code import print_code

    import_code = get_import_module_code(str(choice_file))
    import_code_robust = lambda_to_python_script(
        lambda: wrap_import_in_try_except(
            import_line=import_code, pyfile=str(choice_file), repo_root=str(repo_root) if repo_root is not None else None
        ),
        in_global=True,
        import_module=False,
    )
    code_printing = lambda_to_python_script(
        lambda: print_code(code=import_code_robust, lexer="python", desc="import as module code"),
        in_global=True, import_module=False
    )
    print(f"🧩 Preparing import code for module import:\n{import_code}")
    if choice_function is not None:
        calling = f"""res = {choice_function}({("**" + str(kwargs_dict)) if kwargs_dict else ""})"""
    else:
        calling = """# No function selected to call. You can add your code here."""
    new_choice_file = Path.home().joinpath(f"tmp_results/tmp_scripts/python/{Path(choice_file).parent.name}_{Path(choice_file).stem}_{randstr_func(10)}.py")
    new_choice_file.parent.mkdir(parents=True, exist_ok=True)
    new_choice_file.write_text(import_code_robust + "\n" + code_printing + "\n" + calling, encoding="utf-8")
    return new_choice_file


def _build_final_command(args: FireJobArgs, exe_line: str, choice_file: Path, choice_file_adjusted: str, choice_function: Optional[str], fire_args: str) -> str:
    """Build the final command string."""
    if args.debug:
        import platform
        if platform.system() == "Windows":
            return f"{exe_line} -m ipdb {choice_file_adjusted} "
        elif platform.system() in ["Linux", "Darwin"]:
            return f"{exe_line} -m pudb {choice_file_adjusted} "
        else:
            raise NotImplementedError(f"Platform {platform.system()} not supported.")
    elif args.module:
        return f"{exe_line} {choice_file_adjusted} "
    elif choice_function is not None and choice_file.suffix == ".py":
        return f"{exe_line} -m fire {choice_file_adjusted} {choice_function} {fire_args}"
    elif args.streamlit:
        if args.holdDirectory:
            return f"{exe_line} {choice_file}"
        else:
            return f"cd {choice_file.parent}\n{exe_line} {choice_file.name}\ncd {Path.cwd()}"
    elif args.cmd:
        return rf""" cd /d {choice_file.parent} & {exe_line} {choice_file.name} """
    elif choice_file.suffix == "":
        return f"{exe_line} {choice_file} {fire_args}"
    else:
        return f"{exe_line} {choice_file} "


def _apply_command_modifiers(args: FireJobArgs, command: str, choice_file: Path, repo_root: Optional[Path], randstr_func: RandStrFunc) -> str:
    """Apply various command modifiers based on args."""
    from rich.panel import Panel
    from rich.console import Console
    from rich.syntax import Syntax

    console = Console()

    if args.cmd:
        new_line = "\n"
        command = rf"""start cmd -Argument "/k {command.replace(new_line, " & ")} " """

    if args.submit_to_cloud:
        command = f"""uv run python -m machineconfig.cluster.templates.cli_click --file {choice_file} """
        if args.function is not None:
            command += f"--function {args.function} "

    if args.optimized:
        command = command.replace("python ", "python -OO ")

    if args.zellij_tab is not None:
        comman_path__ = Path.home().joinpath(f"tmp_results/tmp_scripts/zellij_commands/{choice_file.stem}_{randstr_func(10)}.sh")
        comman_path__.parent.mkdir(parents=True, exist_ok=True)
        comman_path__.write_text(command, encoding="utf-8")
        console.print(Panel(Syntax(command, lexer="shell"), title=f"🔥 fire command @ {comman_path__}: "), style="bold red")
        import subprocess

        existing_tab_names = subprocess.run(["zellij", "action", "query-tab-names"], capture_output=True, text=True, check=True).stdout.splitlines()
        if args.zellij_tab in existing_tab_names:
            print(f"⚠️ Tab name `{args.zellij_tab}` already exists. Please choose a different name.")
            args.zellij_tab += f"_{randstr_func(3)}"
        from machineconfig.cluster.sessions_managers.zellij_local import run_command_in_zellij_tab

        command = run_command_in_zellij_tab(command=str(comman_path__), tab_name=args.zellij_tab, cwd=None)

    if args.watch:
        command = "watchexec --restart --exts py,sh,ps1 " + command

    if args.git_pull:
        command = f"\ngit -C {choice_file.parent} pull\n" + command

    if args.PathExport:
        from machineconfig.scripts.python.helpers.helpers_fire_command.file_wrangler import add_to_path
        export_line = add_to_path(path_variable="PYTHONPATH", directory=str(repo_root))
        command = export_line + "\n" + command

    if args.loop:
        import platform
        if platform.system() in ["Linux", "Darwin"]:
            command = command + "\nsleep 0.5"
        elif platform.system() == "Windows":
            command = "$ErrorActionPreference = 'SilentlyContinue';\n" + command + "\nStart-Sleep -Seconds 0.5"
        else:
            raise NotImplementedError(f"Platform {platform.system()} not supported.")

    return command



"""ftpx - File transfer utility through SSH."""

import typer
from typing import Annotated


def ftpx(
    source: Annotated[str, typer.Argument(help="Source path (machine:path)")],
    target: Annotated[str, typer.Argument(help="Target path (machine:path)")],
    recursive: Annotated[bool, typer.Option("--recursive", "-r", help="Send recursively.")] = False,
    zipFirst: Annotated[bool, typer.Option("--zipFirst", "-z", help="Zip before sending.")] = False,
    cloud: Annotated[bool, typer.Option("--cloud", "-c", help="Transfer through the cloud.")] = False,
    overwrite_existing: Annotated[bool, typer.Option("--overwrite-existing", "-o", help="Overwrite existing files on remote when sending from local to remote.")] = False,
) -> None:
    """File transfer utility through SSH."""
    from machineconfig.scripts.python.helpers.helpers_network.ftpx_impl import ftpx as impl
    impl(source=source, target=target, recursive=recursive, zipFirst=zipFirst, cloud=cloud, overwrite_existing=overwrite_existing)


def main() -> None:
    """Entry point function that uses typer to parse arguments and call main."""
    app = typer.Typer()
    app.command(no_args_is_help=True, help=ftpx.__doc__, short_help="File transfer utility through SSH.")(ftpx)
    app()


if __name__ == "__main__":
    main()





from typing import Literal


def generate_qrcode_grid(
    strings: list[str],
    output_path: str,
    per_row: int = 3,
    qr_size: int = 200,
    label_height: int = 30,
    padding: int = 20,
    label_max_chars: int = 25,
    format: Literal["svg", "png"] = "svg",
) -> str:
    from pathlib import Path
    if not strings:
        raise ValueError("strings list cannot be empty")

    output_path_obj = Path(output_path)
    output_path_obj.parent.mkdir(parents=True, exist_ok=True)

    if format == "svg":
        return _generate_svg(strings, output_path, per_row, qr_size, label_height, padding, label_max_chars)
    elif format == "png":
        return _generate_png(strings, output_path, per_row, qr_size, label_height, padding, label_max_chars)
    else:
        raise ValueError(f"Unsupported format: {format}")


def _generate_svg(
    strings: list[str],
    output_path: str,
    per_row: int,
    qr_size: int,
    label_height: int,
    padding: int,
    label_max_chars: int,
) -> str:
    num_items = len(strings)
    num_rows = (num_items + per_row - 1) // per_row

    cell_width = qr_size
    cell_height = qr_size + label_height
    total_width = per_row * cell_width + (per_row + 1) * padding
    total_height = num_rows * cell_height + (num_rows + 1) * padding

    from xml.etree import ElementTree as ET

    import qrcode
    svg_root = ET.Element(
        "svg",
        xmlns="http://www.w3.org/2000/svg",
        width=str(total_width),
        height=str(total_height),
        viewBox=f"0 0 {total_width} {total_height}",
    )

    _bg_rect = ET.SubElement(svg_root, "rect", width=str(total_width), height=str(total_height), fill="white")

    for idx, text in enumerate(strings):
        row = idx // per_row
        col = idx % per_row

        x_offset = padding + col * (cell_width + padding)
        y_offset = padding + row * (cell_height + padding)

        qr = qrcode.QRCode(version=1, error_correction=qrcode.ERROR_CORRECT_L, box_size=10, border=2, image_factory=qrcode.image.svg.SvgPathImage)  # type: ignore
        qr.add_data(text)
        qr.make(fit=True)

        qr_img = qr.make_image()
        qr_svg_string = qr_img.to_string(encoding="unicode")

        qr_tree = ET.fromstring(qr_svg_string)

        group = ET.SubElement(svg_root, "g", transform=f"translate({x_offset}, {y_offset})")

        qr_group = ET.SubElement(group, "g")
        for child in qr_tree:
            qr_group.append(child)

        label_text = text[:label_max_chars] if len(text) > label_max_chars else text
        text_y = qr_size + label_height // 2

        text_elem = ET.SubElement(
            group,
            "text",
            x=str(qr_size // 2),
            y=str(text_y),
            fill="black",
            attrib={
                "font-family": "monospace",
                "font-size": "12",
                "text-anchor": "middle",
                "dominant-baseline": "middle",
            },
        )
        text_elem.text = label_text

    tree = ET.ElementTree(svg_root)
    ET.indent(tree, space="  ")
    tree.write(output_path, encoding="unicode", xml_declaration=True)

    return output_path


def _generate_png(
    strings: list[str],
    output_path: str,
    per_row: int,
    qr_size: int,
    label_height: int,
    padding: int,
    label_max_chars: int,
) -> str:
    num_items = len(strings)
    num_rows = (num_items + per_row - 1) // per_row

    cell_width = qr_size
    cell_height = qr_size + label_height
    total_width = per_row * cell_width + (per_row + 1) * padding
    total_height = num_rows * cell_height + (num_rows + 1) * padding

    import qrcode
    import qrcode.image.pil
    from PIL import Image, ImageDraw, ImageFont
    img = Image.new("RGB", (total_width, total_height), color="white")
    draw = ImageDraw.Draw(img)

    try:
        font = ImageFont.truetype("/usr/share/fonts/truetype/dejavu/DejaVuSansMono.ttf", 12)
    except OSError:
        try:
            font = ImageFont.truetype("arial.ttf", 12)
        except OSError:
            font = ImageFont.load_default()

    for idx, text in enumerate(strings):
        row = idx // per_row
        col = idx % per_row

        x_offset = padding + col * (cell_width + padding)
        y_offset = padding + row * (cell_height + padding)

        qr = qrcode.QRCode(version=1, error_correction=qrcode.ERROR_CORRECT_L, box_size=10, border=2, image_factory=qrcode.image.pil.PilImage)
        qr.add_data(text)
        qr.make(fit=True)

        qr_img = qr.make_image(fill_color="black", back_color="white")
        qr_img_resized = qr_img.resize((qr_size, qr_size), Image.Resampling.LANCZOS)

        img.paste(qr_img_resized, (x_offset, y_offset))

        label_text = text[:label_max_chars] if len(text) > label_max_chars else text

        bbox = draw.textbbox((0, 0), label_text, font=font)
        text_width = bbox[2] - bbox[0]
        text_x = x_offset + (qr_size - text_width) // 2
        text_y = y_offset + qr_size + label_height // 2 - 6

        draw.text((text_x, text_y), label_text, fill="black", font=font)

    img.save(output_path, format="PNG")

    return output_path




from pathlib import Path
from rich.text import Text
from rich.panel import Panel
from rich.console import Console
import subprocess
from typing import Optional, Union, Iterable, overload, Literal, cast

@overload
def choose_from_options[T](options: Iterable[T], msg: str, multi: Literal[False], custom_input: bool = False, header: str = "", tail: str = "", prompt: str = "", default: Optional[T] = None, tv: bool = False, preview: Optional[Literal["bat"]]=None) -> T: ...
@overload
def choose_from_options[T](options: Iterable[T], msg: str, multi: Literal[True], custom_input: bool = True, header: str = "", tail: str = "", prompt: str = "", default: Optional[T] = None, tv: bool = False, preview: Optional[Literal["bat"]]=None) -> list[T]: ...
def choose_from_options[T](options: Iterable[T], msg: str, multi: bool, custom_input: bool = True, header: str = "", tail: str = "", prompt: str = "", default: Optional[T] = None, tv: bool = False, preview: Optional[Literal["bat"]]=None, ) -> Union[T, list[T]]:
    # TODO: replace with https://github.com/tmbo/questionary
    # # also see https://github.com/charmbracelet/gum
    options_strings: list[str] = [str(x) for x in options]
    default_string = str(default) if default is not None else None
    console = Console()
    from machineconfig.utils.installer_utils.installer_locator_utils import check_tool_exists
    # from machineconfig.utils.installer_utils.installer_cli import check_tool_exists
    # print("ch1")
    if tv and check_tool_exists("tv"):
        # from pyfzf.pyfzf import FzfPrompt
        # fzf_prompt = FzfPrompt()
        # nl = "\n"
        # choice_string_multi: list[str] = fzf_prompt.prompt(choices=options_strings, fzf_options=("--multi" if multi else "") + f' --prompt "{prompt.replace(nl, " ")}" --ansi')  # --border-label={msg.replace(nl, ' ')}")
        # print("ch2")
        from machineconfig.utils.accessories import randstr
        options_txt_path = Path.home().joinpath("tmp_results/tmp_files/choices_" + randstr(6) + ".txt")
        options_txt_path.parent.mkdir(parents=True, exist_ok=True)
        options_txt_path.write_text("\n".join(options_strings), encoding="utf-8")

        # Run `tv` interactively so the user can make selections. We redirect tv's
        # stdout to a temporary output file so we can read the chosen lines after
        # the interactive session completes. Do not capture_output or redirect
        # stdin/stderr here so `tv` stays attached to the terminal.
        tv_out_path = options_txt_path.with_name(options_txt_path.stem + "_out.txt")
        if preview is None:
            preview_line = ""
        elif preview == "bat":
            preview_line = r"""--preview-command "bat -n --color=always {}" --preview-size 70 """
        
        import platform
        if platform.system() == "Windows":
            # Set UTF-8 encoding for PowerShell to handle emojis properly when piping to external programs
            # $OutputEncoding controls what encoding PowerShell uses when piping to external programs
            # [Console]::OutputEncoding controls how PowerShell interprets output from external programs
            tv_cmd = f"""$OutputEncoding = [Console]::InputEncoding = [Console]::OutputEncoding = [System.Text.UTF8Encoding]::new(); Get-Content -Encoding UTF8 {options_txt_path} | tv  {preview_line} --ansi true --source-output "{{strip_ansi}}" | Out-File -Encoding utf8 -FilePath {tv_out_path} """
        else:
            tv_cmd = f"""cat {options_txt_path} | tv  {preview_line} --ansi true --source-output "{{strip_ansi}}" > {tv_out_path} """

        print(f"Running tv command: {tv_cmd}")
        # res = subprocess.run(tv_cmd, shell=True)
        from machineconfig.utils.code import run_shell_script
        res = run_shell_script(tv_cmd, display_script=False, clean_env=False)
        
        # If tv returned a non-zero code and there is no output file, treat it as an error.
        if res.returncode != 0 and not tv_out_path.exists():
            raise RuntimeError(f"Got error running tv command: {tv_cmd}\nreturncode: {res.returncode}")

        # Read selections (if any) from the output file created by tv.
        print(f"Reading tv output from: {tv_out_path}")
        out_text = tv_out_path.read_text(encoding="utf-8-sig")
        choice_string_multi = [x for x in out_text.splitlines() if x.strip() != ""]

        # Cleanup temporary files
        options_txt_path.unlink(missing_ok=True)
        tv_out_path.unlink(missing_ok=True)
        if not multi:
            try:
                choice_one_string = choice_string_multi[0]
                if isinstance(list(options)[0], str): return cast(T, choice_one_string)
                choice_idx = options_strings.index(choice_one_string)
                return list(options)[choice_idx]
            except IndexError as ie:
                print(f"❌ Error: {options=}, {choice_string_multi=}")
                print(f"🔍 Available choices: {choice_string_multi}")
                raise ie
        if isinstance(list(options)[0], str):
            return cast(list[T], choice_string_multi)
        choice_idx_s = [options_strings.index(x) for x in choice_string_multi]
        return [list(options)[x] for x in choice_idx_s]
    else:
        if default is not None:
            assert default in options, f"Default `{default}` option not in options `{list(options)}`"
            default_msg = Text(" <<<<-------- DEFAULT", style="bold red")
        else:
            default_msg = Text("")
        txt = Text("\n" + msg + "\n")
        for idx, key in enumerate(options):
            txt = txt + Text(f"{idx:2d} ", style="bold blue") + str(key) + (default_msg if default is not None and default == key else "") + "\n"
        txt_panel = Panel(txt, title=header, subtitle=tail, border_style="bold red")

        console.print(txt_panel)
        if default is not None:
            choice_string = input(f"{prompt}\nEnter option number/name or hit enter for default choice: ")
        else:
            choice_string = input(f"{prompt}\nEnter option number/name: ")

        if choice_string == "":
            if default_string is None:
                console.print(Panel("🧨 Default option not available!", title="Error", expand=False))
                return choose_from_options(msg=msg, options=options, header=header, tail=tail, prompt=prompt, default=default, tv=tv, multi=multi, custom_input=custom_input)
            choice_idx = options_strings.index(default_string)
            assert default is not None, "🧨 Default option not available!"
            choice_one: T = default
        else:
            try:
                choice_idx = int(choice_string, base=10)
                choice_one = list(options)[choice_idx]
            except IndexError as ie:  # i.e. converting to integer was successful but indexing failed.
                if choice_string in options_strings:  # string input
                    choice_idx = options_strings.index(choice_one)  # type: ignore
                    choice_one = list(options)[choice_idx]
                elif custom_input:
                    return str(choice_string)  # type: ignore
                else:
                    _ = ie
                    # raise ValueError(f"Unknown choice. {choice_string}") from ie
                    console.print(Panel(f"❓ Unknown choice: '{choice_string}'", title="Error", expand=False))
                    return choose_from_options(msg=msg, options=options, header=header, tail=tail, prompt=prompt, default=default, tv=tv, multi=multi, custom_input=custom_input)
            except (TypeError, ValueError) as te:  # int(choice_string) failed due to # either the number is invalid, or the input is custom.
                if choice_string in options_strings:  # string input
                    choice_idx = options_strings.index(choice_one)  # type: ignore
                    choice_one = list(options)[choice_idx]
                elif custom_input:
                    return choice_string  # type: ignore
                else:
                    _ = te
                    # raise ValueError(f"Unknown choice. {choice_string}") from te
                    console.print(Panel(f"❓ Unknown choice: '{choice_string}'", title="Error", expand=False))
                    return choose_from_options(msg=msg, options=options, header=header, tail=tail, prompt=prompt, default=default, tv=tv, multi=multi, custom_input=custom_input)
        console.print(Panel(f"✅ Selected option {choice_idx}: {choice_one}", title="Selected", expand=False))
        if multi:
            return [choice_one]
    return choice_one


def choose_cloud_interactively() -> str:
    console = Console()
    console.print(Panel("🔍 LISTING CLOUD REMOTES | Fetching available cloud remotes...", border_style="bold blue", expand=False))
    result = subprocess.run("rclone listremotes", shell=True, capture_output=True, text=True)
    tmp = result.stdout if result.returncode == 0 else None
    if isinstance(tmp, str):
        remotes: list[str] = [x.replace(":", "") for x in tmp.splitlines()]

    else:
        raise ValueError(f"Got {tmp} from rclone listremotes")
    if len(remotes) == 0:
        raise RuntimeError("You don't have remotes. Configure your rclone first to get cloud services access.")
    cloud: str = choose_from_options(msg="WHICH CLOUD?", multi=False, options=list(remotes), default=remotes[0], tv=True)
    console.print(Panel(f"✅ SELECTED CLOUD | {cloud}", border_style="bold blue", expand=False))
    return cloud


def get_ssh_hosts() -> list[str]:
    from paramiko import SSHConfig

    c = SSHConfig()
    c.parse(open(Path.home().joinpath(".ssh/config"), encoding="utf-8"))
    return list(c.get_hostnames())


@overload
def choose_ssh_host(multi: Literal[False]) -> str: ...
@overload
def choose_ssh_host(multi: Literal[True]) -> list[str]: ...
def choose_ssh_host(multi: bool):
    return choose_from_options(msg="", options=get_ssh_hosts(), multi=multi, tv=True)




from pathlib import Path
from machineconfig.utils.accessories import randstr
from machineconfig.utils.meta import lambda_to_python_script
from machineconfig.utils.ssh_utils.abc import MACHINECONFIG_VERSION, DEFAULT_PICKLE_SUBDIR
from machineconfig.utils.code import get_uv_command
from typing import Union


def create_dir_and_check_if_exists(self: "SSH", path_rel2home: str, overwrite_existing: bool) -> None:
    """Helper to create a directory on remote machine and return its path."""

    def create_target_dir(target_rel2home: str, overwrite: bool):
        from pathlib import Path
        import shutil
        target_path_abs = Path(target_rel2home).expanduser()
        if not target_path_abs.is_absolute():
            target_path_abs = Path.home().joinpath(target_path_abs)
        if overwrite and target_path_abs.exists():
            if str(target_path_abs) == str(Path.home()):
                raise RuntimeError("Refusing to overwrite home directory!")
            if target_path_abs.is_dir():
                shutil.rmtree(target_path_abs)
            else:
                target_path_abs.unlink()
        print(f"Creating directory for path: {target_path_abs}")
        target_path_abs.parent.mkdir(parents=True, exist_ok=True)
    command = lambda_to_python_script(
        lambda: create_target_dir(target_rel2home=path_rel2home, overwrite=overwrite_existing),
        in_global=True, import_module=False
    )
    tmp_py_file = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/create_target_dir_{randstr()}.py")
    tmp_py_file.parent.mkdir(parents=True, exist_ok=True)
    tmp_py_file.write_text(command, encoding="utf-8")
    assert self.sftp is not None
    tmp_remote_path = ".tmp_pyfile.py"
    self.sftp.put(localpath=str(tmp_py_file), remotepath=str(Path(self.remote_specs["home_dir"]).joinpath(tmp_remote_path)))
    resp = self.run_shell_cmd_on_remote(
        command=f"""{get_uv_command(platform=self.remote_specs['system'])} run python {tmp_remote_path}""",
        verbose_output=False,
        description=f"Creating target dir {path_rel2home}",
        strict_stderr=True,
        strict_return_code=True,
    )
    resp.print(desc=f"Created target dir {path_rel2home}")


def check_remote_is_dir(self: "SSH", source_path: Union[str, Path]) -> bool:
    """Helper to check if a remote path is a directory."""

    def check_is_dir(path_to_check: str, json_output_path: str) -> bool:
        from pathlib import Path
        import json

        is_directory = Path(path_to_check).expanduser().absolute().is_dir()
        json_result_path = Path(json_output_path)
        json_result_path.parent.mkdir(parents=True, exist_ok=True)
        json_result_path.write_text(json.dumps(is_directory, indent=2), encoding="utf-8")
        print(json_result_path.as_posix())
        return is_directory

    remote_json_output = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/return_{randstr()}.json").as_posix()
    command = lambda_to_python_script(
        lambda: check_is_dir(path_to_check=str(source_path), json_output_path=remote_json_output),
        in_global=True, import_module=False
    )
    response = self.run_py_remotely(
        python_code=command,
        uv_with=[MACHINECONFIG_VERSION],
        uv_project_dir=None,
        description=f"Check if source `{source_path}` is a dir",
        verbose_output=False,
        strict_stderr=False,
        strict_return_code=False,
    )
    remote_json_path = response.op.strip()
    if not remote_json_path:
        raise RuntimeError(f"Failed to check if {source_path} is directory - no response from remote")

    local_json = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/local_{randstr()}.json")
    self.simple_sftp_get(remote_path=remote_json_path, local_path=local_json)
    import json

    try:
        result = json.loads(local_json.read_text(encoding="utf-8"))
    except (json.JSONDecodeError, FileNotFoundError) as err:
        raise RuntimeError(f"Failed to check if {source_path} is directory - invalid JSON response: {err}") from err
    finally:
        if local_json.exists():
            local_json.unlink()
    assert isinstance(result, bool), f"Failed to check if {source_path} is directory"
    return result

def expand_remote_path(self: "SSH", source_path: Union[str, Path]) -> str:
    """Helper to expand a path on the remote machine."""

    def expand_source(path_to_expand: str, json_output_path: str) -> str:
        from pathlib import Path
        import json

        expanded_path_posix = Path(path_to_expand).expanduser().absolute().as_posix()
        json_result_path = Path(json_output_path)
        json_result_path.parent.mkdir(parents=True, exist_ok=True)
        json_result_path.write_text(json.dumps(expanded_path_posix, indent=2), encoding="utf-8")
        print(json_result_path.as_posix())
        return expanded_path_posix

    remote_json_output = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/return_{randstr()}.json").as_posix()
    command = lambda_to_python_script(
        lambda: expand_source(path_to_expand=str(source_path), json_output_path=remote_json_output),
        in_global=True, import_module=False
    )
    response = self.run_py_remotely(
        python_code=command,
        uv_with=[MACHINECONFIG_VERSION],
        uv_project_dir=None,
        description="Resolving source path by expanding user",
        verbose_output=False,
        strict_stderr=False,
        strict_return_code=False,
    )
    remote_json_path = response.op.strip()
    if not remote_json_path:
        raise RuntimeError(f"Could not resolve source path {source_path} - no response from remote")

    local_json = Path.home().joinpath(f"{DEFAULT_PICKLE_SUBDIR}/local_{randstr()}.json")
    self.simple_sftp_get(remote_path=remote_json_path, local_path=local_json)
    import json

    try:
        result = json.loads(local_json.read_text(encoding="utf-8"))
    except (json.JSONDecodeError, FileNotFoundError) as err:
        raise RuntimeError(f"Could not resolve source path {source_path} - invalid JSON response: {err}") from err
    finally:
        if local_json.exists():
            local_json.unlink()
    assert isinstance(result, str), f"Could not resolve source path {source_path}"
    return result


if __name__ == "__main__":
    from machineconfig.utils.ssh import SSH
